from snsynth.transform.definitions import ColumnType
from .base import ColumnTransformer
import numpy as np

class LogTransformer(ColumnTransformer):
    """Logarithmic transformation of values.  Useful for transforming skewed data.

    """
    def __init__(self):
        super().__init__()
    @property
    def output_type(self):
        return ColumnType.CONTINUOUS
    @property
    def cardinality(self):
        return [None]
    def _fit(self, val, idx=None):
        pass
    def _clear_fit(self):
        # this transform doesn't need fit
        self._fit_complete = True
        self.output_width = 1
    def _transform(self, val):
        if val is None or (isinstance(val, float) and np.isnan(val)):
            return None
        else:
            return float(np.log(val))
    def _inverse_transform(self, val):
        if val is None or (isinstance(val, float) and np.isnan(val)):
            return np.nan
        else:
            return float(np.exp(val))


import pandas as pd
import numpy as np
import warnings
import re
from snsynth import transform
from snsynth.transform.anonymization import AnonymizationTransformer
from snsynth.transform.base import ColumnTransformer
from snsynth.transform.drop import DropTransformer
from snsynth.transform.minmax import MinMaxTransformer
from snsynth.transform.bin import BinTransformer
from snsynth.transform.label import LabelTransformer
from snsynth.transform.onehot import OneHotEncoder
from snsynth.transform.chain import ChainTransformer
from snsynth.transform.datetime import DateTimeTransformer

class SequenceCounter:
    def __init__(self):
        self.count = -1
    def __call__(self, *args, **kwargs):
        self.count += 1
        return self.count

_EXPECTED_COL_STYLES = {
    'gan': {
        'categorical': [OneHotEncoder],
        'ordinal': [OneHotEncoder],
        'continuous': [MinMaxTransformer],
    },
    'cube': {
        'categorical': [LabelTransformer],
        'ordinal': [LabelTransformer],
        'continuous': [BinTransformer],
    }
}

class TypeMap:
    def __init__(self):
        pass    
    @classmethod
    def get_transformers(cls, column_names, style='gan', *ignore, nullable=False, categorical_columns=[], ordinal_columns=[], continuous_columns=[], special_types={}, constraints=None):
        if ordinal_columns is None:
            ordinal_columns = []
        if continuous_columns is None:
            continuous_columns = []
        if categorical_columns is None:
            categorical_columns = []
        if constraints is None:
            constraints = dict()

        constrained_columns = set(constraints.keys())
        for k, v in constraints.items(): # check if a string value is an alias for any of the lists
            if isinstance(v, str) and v.lower() in {"categorical", "ordinal", "continuous"}:
                if v.lower() == "categorical":
                  categorical_columns.append(k)
                elif v.lower() == "ordinal":
                   ordinal_columns.append(k)
                elif v.lower() == "continuous":
                    continuous_columns.append(k)
                constrained_columns.remove(k)

        transformers = []
        for col in list(column_names):
            if col in constrained_columns:
                transformer = cls._get_transformer(constraints[col])
                transformers.append(transformer)
            elif col in special_types and special_types[col] is not None:
                stype = special_types[col]
                if stype in ['email', 'ssn', 'uuid4']:
                    transformers.append(AnonymizationTransformer(stype))
                elif stype == 'sequence':
                    transformers.append(AnonymizationTransformer(SequenceCounter()))
                elif stype == 'datetime':
                    if style == 'gan':
                        t = ChainTransformer([
                            DateTimeTransformer(),
                            MinMaxTransformer(nullable=nullable)
                        ])
                        transformers.append(t)
                    elif style == 'cube':
                        t = ChainTransformer([
                            DateTimeTransformer(),
                            BinTransformer(bins=20, nullable=nullable)
                        ])
                        transformers.append(t)
                    else:
                        raise ValueError(f"Unknown style: {style}")
                else:
                    raise ValueError(f"Unknown special type {stype}")
            elif col in categorical_columns:
                if style == 'gan':
                    t = ChainTransformer([LabelTransformer(nullable=nullable), OneHotEncoder()])
                    transformers.append(t)
                elif style == 'cube':
                    t = LabelTransformer(nullable=nullable)
                    transformers.append(t)
                else:
                    raise ValueError(f"Unknown style: {style}")
            elif col in ordinal_columns:
                if style == 'gan':
                    t = ChainTransformer([LabelTransformer(nullable=nullable), OneHotEncoder()])
                    transformers.append(t)
                elif style == 'cube':
                    t = LabelTransformer(nullable=nullable)
                    transformers.append(t)
                else:
                    raise ValueError(f"Unknown style: {style}")
            elif col in continuous_columns:
                if style == 'gan':
                    t = MinMaxTransformer(nullable=nullable)
                    transformers.append(t)
                elif style == 'cube':
                    t = BinTransformer(nullable=nullable)
                    transformers.append(t)
                else:
                    raise ValueError(f"Unknown style: {style}")
            else:
                raise ValueError(f"Column in dataframe not specified as categorical, ordinal, continuous or constraint: {col}")
        return transformers
    @classmethod
    def infer_column_types(cls, data, excluded_columns=None):
        max_cached = 1000
        n_columns = 0
        colnames = []
        coltypes = []
        nullable = []
        pii = []
    
        if isinstance(data, pd.DataFrame):
            colnames = list(data.columns)
            n_columns = len(colnames)
            data = [tuple([c for c in t[1:]]) for t in data.itertuples()]
        elif isinstance(data, list):
            colnames = [v for v in data[0]]
            colname_types = set([type(v) for v in colnames])
            if len(colname_types) != 1 or str not in colname_types:
                colnames = [i for i in range(len(colnames))]
            n_columns = len(colnames)
        elif isinstance(data, np.ndarray):
            n_columns = data.shape[1]
            colnames = [i for i in range(n_columns)]
            data = data.tolist()

        excluded_indices = set()
        if excluded_columns is not None and len(excluded_columns) > 0: # don't infer the excluded columns
            colnames_actual = []
            for i, c in enumerate(colnames):
                if c in excluded_columns:
                    excluded_indices.add(i)
                else:
                    colnames_actual.append(c)
            colnames = colnames_actual
            n_columns = len(colnames)

        # cache up to max_cached rows
        n_cached = 0
        value_cache = []
        for _ in range(n_columns):
            value_cache.append([])
            pii.append(None)
            nullable.append(False)
            coltypes.append(None)
        for row in data:
            index = 0
            for i, val in enumerate(row):
                if i not in excluded_indices:
                    value_cache[index].append(val)
                    index += 1
            n_cached += 1
            if n_cached >= max_cached:
                break

        # infer each column type
        for i in range(n_columns):
            # check for nulls
            if any([v is None or isinstance(v, float) and np.isnan(v) for v in value_cache[i]]):
                nullable[i] = True
            # remove nulls
            value_cache[i] = [v for v in value_cache[i] if v is not None and not (isinstance(v, float) and np.isnan(v))]
            # check for bools
            if any([isinstance(v, bool) for v in value_cache[i]]):
                coltypes[i] = 'categorical'
            # check for strings
            elif any([isinstance(v, str) for v in value_cache[i]]):
                # check for pii
                pii_type = cls.infer_pii(value_cache[i])
                if pii_type is not None:
                    pii[i] = pii_type
                else:
                    distinct = set(value_cache[i])
                    if len(distinct) > 0.8 * len(value_cache[i]):
                        warnings.warn(f"Column {colnames[i]} has {len(distinct)} distinct values out of {len(value_cache[i])} total values. This appears to be unbounded categorical data and may risk privacy leaks.")
                coltypes[i] = 'categorical'
            elif any([isinstance(v, int) for v in value_cache[i]]):
                distinct = set(value_cache[i])
                if len(distinct) < 150 and max(distinct) - min(distinct) < 150:
                    coltypes[i] ='ordinal'
                else:
                    # check for sequence
                    if cls.is_ids(value_cache[i]):
                        pii[i] = 'sequence'
                    coltypes[i] = 'continuous'
            elif any([isinstance(v, float) for v in value_cache[i]]):
                if all([v.is_integer() for v in value_cache[i]]):
                    distinct = set(value_cache[i])
                    threshold = n_cached * 0.20 # 20% of rows
                    if len(distinct) < threshold and max(distinct) - min(distinct) < threshold:
                        coltypes[i] = 'ordinal'
                    else:
                        # check for sequence
                        if cls.is_ids(value_cache[i]):
                            pii[i] = 'sequence'
                        coltypes[i] = 'continuous'
                else:
                    coltypes[i] = 'continuous'
            else:
                distinct = set(value_cache[i])
                threshold = n_cached * 0.20 # 20% of rows
                if len(distinct) < threshold:
                    coltypes[i] = 'categorical'
                else:
                    raise ValueError(f"Cannot infer a column type for column {i}")
        
        result = {
            'columns': colnames,
            'categorical_columns': [colnames[i] for i, v in enumerate(coltypes) if v == 'categorical'],
            'ordinal_columns': [colnames[i] for i, v in enumerate(coltypes) if v == 'ordinal'],
            'continuous_columns': [colnames[i] for i, v in enumerate(coltypes) if v == 'continuous'],
            'nullable_columns': [colnames[i] for i, v in enumerate(nullable) if v],
            'pii': pii
        }
        return result

    @classmethod
    def infer_pii(cls, vals):
        uuid4string = re.compile(r'[0-9a-f]{8}-?[0-9a-f]{4}-?4[0-9a-f]{3}-?[89ab][0-9a-f]{3}-?[0-9a-f]{12}')
        email = re.compile(r'[^@]+@[^@]+\.[^@]+')
        ssn = re.compile(r'^\d{3}-?\d{2}-?\d{4}$')
        dt = DateTimeTransformer()
        types = []
        for v in vals:
            try:
                parsed = dt._parse_date(v)
                if parsed is not None:
                    types.append('datetime')
            except:
                pass
            if uuid4string.match(v):
                types.append('uuid4')
            elif email.match(v):
                types.append('email')
            elif ssn.match(v):
                types.append('ssn')
        if len(types) == 0:
            return None
        types = list(set(types))
        if len(types) == 1:
            return types[0]
        else:
            warnings.warn(f"Multiple PII types detected: {types}")
            return types[0]

    @classmethod
    def is_ids(cls, vals):
        # must pass in ints or floats
        distinct = set(vals)
        if len(distinct) < 0.8 * len(vals):
            return False
        _range = max(distinct) - min(distinct)
        if _range >= len(distinct) and _range < 1.1 * len(distinct):
            return True
        vals = sorted(list(distinct))
        diffs = [vals[i+1] - vals[i] for i in range(len(vals)-1)]
        if len(set(diffs)) < 4:
            return True
        if np.std(diffs) < 0.1 * np.mean(diffs):
            return True
        return False

    @staticmethod
    def _get_transformer(constraint):
        """
        Constructs a transformer object according to the given constraint.
        Firstly, if the string is a keyword e.g. 'drop' the corresponding transformer is returned.
        Secondly, the method tries to construct a transformer class from a string value.
        Other values will be passed into an ``AnonymizationTransformer``.

        :param constraint: Transformer to be used, a keyword / built-in Faker method
        :type constraint: a ColumnTransformer object/type, a string, or a callable
        :return: Transformer object
        :rtype: ColumnTransformer
        """
        if isinstance(constraint, ColumnTransformer):
            return constraint
        elif isinstance(constraint, type) and issubclass(constraint, ColumnTransformer):
            return constraint()
        elif isinstance(constraint, str):
            # firstly, check if column should be ignored completely
            if constraint.lower() == "drop":
                return DropTransformer()

            # secondly, try to find a valid transformer class
            try:
                transformer_class = getattr(transform, constraint)
                return transformer_class()
            except AttributeError:
                pass

        # otherwise assume that e.g. a Faker method is given
        try:
            return AnonymizationTransformer(constraint)
        except ValueError as e:
            raise ValueError(
                f"Provided `constraint` {constraint} is invalid. Please read the documentation about constraints."
            ) from e



import logging
import warnings

import numpy as np
import pandas as pd

from snsynth.base import Synthesizer

logger = logging.getLogger(__name__)


class QUAILSynthesizer(Synthesizer):
    """
    Quailified Architecture to Improve Labeling.
    Divide epsilon in a known classification task
    between a differentially private synthesizer and
    classifier. Train DP classifier on real, fit DP synthesizer
    to features (excluding the target label),
    and use synthetic data from the DP synthesizer with
    the DP classifier to create artificial labels. Produces
    complete synthetic data.

    More information here:
    Differentially Private Synthetic Data: Applied Evaluations and Enhancements
    https://arxiv.org/abs/2011.05537

    :param epsilon: Total epsilon used across the DP Synthesizer and DP Classifier
    :type epsilon: float
    :param dp_synthesizer: A function that returns an instance of a DP Synthesizer
        for a specified epsilon value
    :type dp_synthesizer: function (epsilon) -> SDGYMBaseSynthesizer
    :param dp_classifier: A function that returns an instance of a DP Classifier
        for a specified epsilon value
    :type dp_classifier: function (epsilon) -> classifier
    :param target: The column name of the target column
    :type target: str
    :param test_size: Percent of the data that should be used for the test set,
        defaults to 0.2
    :type test_size: float, optional
    :param seed: Seed for controlling randomness for testing, defaults to None
    :type seed: int, optional
    :param eps_split: Percent of epsilon used for the classifier.
        1 - eps_split is used for the Synthesizer., defaults to 0.9
    :type eps_split: float, optional
    """
    def __init__(
        self,
        epsilon,
        dp_synthesizer,
        dp_classifier,
        target,
        test_size=0.2,
        seed=None,
        eps_split=0.9,
    ):
        """
        Quailified Architecture to Improve Labeling.
        Divide epsilon in a known classification task
        between a differentially private synthesizer and
        classifier. Train DP classifier on real, fit DP synthesizer
        to features (excluding the target label),
        and use synthetic data from the DP synthesizer with
        the DP classifier to create artificial labels. Produces
        complete synthetic data.

        More information here:
        Differentially Private Synthetic Data: Applied Evaluations and Enhancements
        https://arxiv.org/abs/2011.05537

        :param epsilon: Total epsilon used across the DP Synthesizer and DP Classifier
        :type epsilon: float
        :param dp_synthesizer: A function that returns an instance of a DP Synthesizer
            for a specified epsilon value
        :type dp_synthesizer: function (epsilon) -> SDGYMBaseSynthesizer
        :param dp_classifier: A function that returns an instance of a DP Classifier
            for a specified epsilon value
        :type dp_classifier: function (epsilon) -> classifier
        :param target: The column name of the target column
        :type target: str
        :param test_size: Percent of the data that should be used for the test set,
            defaults to 0.2
        :type test_size: float, optional
        :param seed: Seed for controlling randomness for testing, defaults to None
        :type seed: int, optional
        :param eps_split: Percent of epsilon used for the classifier.
            1 - eps_split is used for the Synthesizer., defaults to 0.9
        :type eps_split: float, optional
        """
        self.epsilon = epsilon
        self.eps_split = eps_split
        self.dp_synthesizer = dp_synthesizer
        self.dp_classifier = dp_classifier
        self.target = target
        self.test_size = test_size
        self.seed = seed

        # Model
        self.private_model = None
        self.private_synth = None

        # Pandas check
        self.pandas = False
        self.pd_cols = None
        self.pd_index = None

    def fit(
        self,
        data,
        categorical_columns=tuple(),
        ordinal_columns=tuple(),
        transformer=None,
        continuous_columns=None,
        verbose=None,
        preprocessor_eps=0.0,
        nullable=False,
    ):
        """
        Takes a dataset and fits the synthesizer/learning model to it, using the epsilon split
        specified in the init.

        :param data: Data
        :type data: pd.DataFrame or np.array
        """
        if verbose is not None:
            warnings.warn("verbose is deprecated. Use logging.setLevel instead")

        from sklearn.model_selection import train_test_split
        from sklearn.metrics import classification_report
        from sklearn.metrics import accuracy_score

        if isinstance(data, pd.DataFrame):
            self.pandas = True
            for col in data.columns:
                data[col] = pd.to_numeric(data[col], errors="ignore")
            self.data = data
            self.pd_cols = data.columns
            self.pd_index = data.index
        else:
            raise ("Only pandas dataframes for data as of now.")

        private_features = data.loc[:, data.columns != self.target]
        private_target = data.loc[:, data.columns == self.target]
        x_train, x_test, y_train, y_test = train_test_split(
            private_features,
            private_target,
            test_size=self.test_size,
            random_state=self.seed,
        )

        # Here we train a differentially private model on the real
        # data. We report on the accuracy for now to give a sense of
        # the upper bound on performance in the sampling step.
        self.private_model = self.dp_classifier(epsilon=(self.epsilon * self.eps_split))
        self.private_model.fit(x_train, y_train.values.ravel())
        predictions = self.private_model.predict(x_test)
        self.class_report = classification_report(
            np.ravel(y_test), predictions, labels=np.unique(predictions)
        )
        self.target_accuracy = accuracy_score(np.ravel(y_test), predictions)
        log_level = logger.level
        if verbose:
            log_level = logging.INFO

        logging.log(log_level, "Internal model report: ")
        logging.log(log_level, self.class_report)
        logging.log(log_level, self.target_accuracy)

        # We use the features in our synthesis.
        self.private_synth = self.dp_synthesizer(
            epsilon=(self.epsilon * (1 - self.eps_split))
        )
        self.private_synth.fit(
            data=private_features,
            categorical_columns=categorical_columns,
            ordinal_columns=ordinal_columns,
            transformer=transformer,
            continuous_columns=continuous_columns,
            preprocessor_eps=preprocessor_eps,
            nullable=nullable,
        )

        if hasattr(self.private_model, "coef_"):
            logging.log(log_level, self.private_model.coef_)

        if hasattr(self.private_model, "intercept_"):
            logging.log(log_level, self.private_model.intercept_)

        if hasattr(self.private_model, "classes_"):
            logging.log(log_level, self.private_model.classes_)

    def sample(self, samples):
        """
        Sample from the synthesizer model.

        :param samples: The number of samples to create
        :type samples: int
        :return: A dataframe of length samples
        :rtype: pd.Dataframe
        """
        sampled_features = self.private_synth.sample(samples)
        y_values = self.private_model.predict(sampled_features)

        sampled_features[self.target] = y_values
        return sampled_features



import pandas as pd
import numpy as np

from pacsynth import (
    DpAggregateSeededSynthesizer,
    DpAggregateSeededParametersBuilder,
    AccuracyMode,
    FabricationMode,
)
from pacsynth import Dataset as AggregateSeededDataset
from snsynth.base import Synthesizer


"""
Wrapper for Private Aggregate Seeded Synthesizer from pac-synth:
https://pypi.org/project/pac-synth/.

A differentially-private synthesizer that relies on DP Marginals to
build synthetic data. It will compute DP Marginals (called aggregates)
for your dataset up to and including a specified reporting length, and
synthesize data based on the computed aggregated counts.

For documentation please refer to:
    - https://github.com/microsoft/synthetic-data-showcase
    - https://github.com/microsoft/synthetic-data-showcase/tree/main/docs/dp
"""


class AggregateSeededSynthesizer(Synthesizer):
    """
    SmartNoise class wrapper for Private Aggregate Seeded Synthesizer from pac-synth.
    Works with Pandas data frames, raw data and follows norms set by other SmartNoise synthesizers.

    :param reporting_length: The maximum length of the combinations to be synthesized.  For example,
        if reporting length is 2, the synthesizer will compute DP marginals for all two-column combinations
        in the dataset.
    :type reporting_length: int
    :param epsilon: The privacy budget to be used for the synthesizer.
    :type epsilon: float
    :param delta: The delta value to be used for the synthesizer.  If set, should be small, in the range
        of 1/(n * sqrt(n)), where n is the approximate number of records in the dataset.
    :param percentile_percentage: Because the synthesizer computes multiple n-way marginals, each individual may
        affect multiple marginals.  The ``percentile_percentage`` can remove the influence of outliers to
        reduce sensitivity and improve the accuracy of the synthesizer.  For example, if ``percentile_percentage``
        is 99, the synthesizer will use a sensitivity that can accomodate 99% of the individuals, and will ensure
        that the records of the outlier 1% are sampled to conform to this sensitivity.
    :type percentile_percentage: int
    :param percentile_epsilon_proportion: The proportion of the epsilon budget to be used to estimate the
        percentile sensitivity.
    :type percentile_epsilon_proportion: float
    :param verbose: Show diagnostic information about the synthesizer's progress.
    :type verbose: bool

    See the `pac-synth documentation <https://github.com/microsoft/synthetic-data-showcase/blob/main/docs/dp/README.md>`_ 
        for more details about these and other hyperparameters.

    Reuses code and modifies it lightly from 
        `pac-synth <https://github.com/microsoft/synthetic-data-showcase/tree/main/packages/lib-pacsynth>`_.
    """

    def __init__(
        self,
        reporting_length=3,
        epsilon=4.0,
        delta=None,
        percentile_percentage=99,
        percentile_epsilon_proportion=0.01,
        accuracy_mode=AccuracyMode.prioritize_long_combinations(),
        number_of_records_epsilon_proportion=0.005,
        fabrication_mode=FabricationMode.uncontrolled(),
        empty_value="",
        use_synthetic_counts=False,
        weight_selection_percentile=95,
        aggregate_counts_scale_factor=None,
        verbose=False
    ):
        """
        Wrapper for Private Aggregate Seeded Synthesizer from pac-synth.

        For more information about the parameters run `help('pacsynth.DpAggregateSeededParametersBuilder')`.
        """
        self.epsilon = epsilon
        self.delta = delta
        self.reporting_length = reporting_length
        self.percentile_percentage = percentile_percentage
        self.percentile_epsilon_proportion = percentile_epsilon_proportion
        self.accuracy_mode = accuracy_mode
        self.number_of_records_epsilon_proportion = number_of_records_epsilon_proportion
        self.fabrication_mode = fabrication_mode
        self.empty_value = empty_value
        self.use_synthetic_counts = use_synthetic_counts
        self.weight_selection_percentile = weight_selection_percentile
        self.aggregate_counts_scale_factor = aggregate_counts_scale_factor
        self.verbose = verbose
        self.preprocessed = False
        self.build_synthesizer()

    def build_synthesizer(self):
        builder = (
            DpAggregateSeededParametersBuilder()
            .reporting_length(self.reporting_length)
            .epsilon(self.epsilon)
            .percentile_percentage(self.percentile_percentage)
            .percentile_epsilon_proportion(self.percentile_epsilon_proportion)
            .accuracy_mode(self.accuracy_mode)
            .number_of_records_epsilon_proportion(self.number_of_records_epsilon_proportion)
            .fabrication_mode(self.fabrication_mode)
            .empty_value(self.empty_value)
            .use_synthetic_counts(self.use_synthetic_counts)
            .weight_selection_percentile(self.weight_selection_percentile)
        )

        if self.aggregate_counts_scale_factor is not None:
            builder = builder.aggregate_counts_scale_factor(
                self.aggregate_counts_scale_factor
            )

        if self.delta is not None:
            builder = builder.delta(self.delta)

        self.reporting_length = self.reporting_length
        self.parameters = builder.build()
        self.synth = DpAggregateSeededSynthesizer(self.parameters)
        self.dataset = None
        self.pandas = False

    def fit(
        self, 
        data,
        *ignore,
        use_columns=None,
        sensitive_zeros=None,
        transformer=None,
        categorical_columns=None,
        ordinal_columns=None,
        continuous_columns=None,
        preprocessor_eps=0.0,
        nullable=False
        ):
        """
        Fit the synthesizer model on the data.

        This will compute the differently private aggregates used to
        synthesize data.

        All the columns are supposed to be categorical, non-categorical columns
        should be binned in advance.

        For more information run `help('pacsynth.Dataset')` and
        `help('pacsynth.DpAggregateSeededSynthesizer.fit')`.

        :param data: The data for fitting the synthesizer model.
        :type data: pd.DataFrame, list[list[str]], AggregateSeededDataset
        :param use_columns: List of column names to be used, defaults to None, meaning use all columns
        :type use_columns: list[str], optional
        :param sensitive_zeros: List of column names containing '0' that should not be turned into empty strings.
        :type sensitive_zeros: list[str], optional
        """

        before_eps = self.epsilon

        train_data = self._get_train_data(
            data,
            style='cube',
            transformer=transformer,
            categorical_columns=categorical_columns, 
            ordinal_columns=ordinal_columns, 
            continuous_columns=continuous_columns, 
            nullable=True,
            preprocessor_eps=preprocessor_eps
        )

        if self.epsilon != before_eps:
            # preprocessor changed epsilon, rebuild synthesizer
            self.build_synthesizer()

        if self._transformer is None:
            raise ValueError("We weren't able to fit a transformer to the data. Please check your data and try again.")

        if self._transformer.output_width > 0:
            colnames = ["column_{}".format(i) for i in range(len(train_data[0]))]
            data = [colnames] + [[str(v) for v in row] for row in train_data]

            assert sensitive_zeros is None, "sensitive zeros cannot be set with a transformer, please set transformer=NoTransformer()"
            assert use_columns is None, "use columns cannot be set with a transformer, please set transformer=NoTransformer()"

            sensitive_zeros = colnames

            cards = self._transformer.cardinality
            if any (c is None for c in cards):
                raise ValueError("The transformer appears to have some continuous columns. Please provide only categorical or ordinal.")

            dimensionality = np.prod(cards)
            if self.verbose:
                print(f"Fitting with {dimensionality} dimensions")


        if isinstance(data, list) and all(map(lambda row: isinstance(row, list), data)):
            self.dataset = AggregateSeededDataset(
                data, use_columns=use_columns, sensitive_zeros=sensitive_zeros
            )
            self.pandas = False
        elif isinstance(data, pd.DataFrame):
            self.dataset = AggregateSeededDataset.from_data_frame(
                data, use_columns=use_columns, sensitive_zeros=sensitive_zeros
            )
            self.pandas = True
        elif isinstance(data, AggregateSeededDataset):
            self.dataset = data
            self.pandas = False
        else:
            raise ValueError(
                "data should be either in raw format (List[List[]]) or a be pandas data frame (pd.DataFrame)"
            )

        self.synth.fit(self.dataset)

    def sample(self, samples=None):
        """
        Sample from the synthesizer model.

        This will sample records from the generated differentially private aggregates.

        If `samples` is `None`, the synthesizer will use all the available differentially
        private attributes counts to synthesize records
        (which will produce a number close to original number of records).

        For more information run `help('pacsynth.DpAggregateSeededSynthesizer.sample')`.

        :param samples: The number of samples to create
        :type samples: int, None
        :return: Generated data samples, the output type adjusts accordingly to the input data.
        :rtype: Dataframe, list[list[str]]
        """
        result = self.synth.sample(samples)

        if self._transformer is not None and self._transformer.output_width > 0:
            result = [[int(v) if v != '' else None for v in row] for row in result[1:]]
            result = self._transformer.inverse_transform(result)
            return result

        if self.pandas is True:
            result = AggregateSeededDataset.raw_data_to_data_frame(result)

        return result


    def get_sensitive_aggregates(
        self, combination_delimiter=";", reporting_length=None
        ):
        """
        Returns the aggregates for the sensitive dataset. For more information run `help('pacsynth.Dataset.get_aggregates')`.

        :param combination_delimiter: Combination delimiter to use, default to ';'
        :type combination_delimiter: str, optional
        :param reporting_length: Maximum length (inclusive) to compute attribute combinations for, defaults to the configured value in the synthesizer
        :type reporting_length: int, optional

        :return: A dictionary with the combination string representation as key and the combination count as value
        :rtype: dict[str, int]
        """
        if self.dataset is None:
            raise RuntimeError(
                "make sure 'fit' method has been successfully called first"
            )

        if reporting_length is None:
            reporting_length = self.reporting_length

        return self.dataset.get_aggregates(reporting_length, combination_delimiter)

    def get_dp_aggregates(self, combination_delimiter=";"):
        """
        Returns the aggregates for the sensitive dataset protected with differential privacy.

        For more information run `help('pacsynth.DpAggregateSeededSynthesizer.get_dp_aggregates')`.

        :param combination_delimiter: Combination delimiter to use, default to ';'
        :type combination_delimiter: str, optional
        :return: A dictionary with the combination string representation as key and the combination count as value
        :rtype: dict[str, int]
        """
        return self.synth.get_dp_aggregates(combination_delimiter)

    def get_dp_number_of_records(self):
        """
        Gets the differentially private number of records computed with the `.fit` method.

        This is different than the number of records specified in the sample method, synthesized
        in the synthetic data. This refers to the differentially private protected number of records
        in original sensitive dataset (Laplacian noise added).

        For more information run `help('pacsynth.DpAggregateSeededSynthesizer.get_dp_number_of_records')`.

        :return: Number of sensitive records protect with differential privacy
        :rtype: int
        """
        return self.synth.get_dp_number_of_records()



import math
import numpy as np
import pandas as pd
import torch
from torch import optim
from torch import nn
import torch.utils.data
from torch.nn import (
    BatchNorm1d,
    Dropout,
    LeakyReLU,
    Linear,
    Module,
    ReLU,
    Sequential,
    Sigmoid,
)
from torch.autograd import Variable
import warnings

from snsynth.base import Synthesizer

from .ctgan.data_sampler import DataSampler
from .ctgan.ctgan import CTGANSynthesizer
from snsynth.transform.table import TableTransformer

from .privacy_utils import weights_init, pate, moments_acc


class Discriminator(Module):
    def __init__(self, input_dim, discriminator_dim, loss, pac=10):
        super(Discriminator, self).__init__()
        torch.cuda.manual_seed(0)
        torch.manual_seed(0)

        dim = input_dim * pac
        #  print ('now dim is {}'.format(dim))
        self.pac = pac
        self.pacdim = dim

        seq = []
        for item in list(discriminator_dim):
            seq += [Linear(dim, item), LeakyReLU(0.2), Dropout(0.5)]
            dim = item

        seq += [Linear(dim, 1)]
        if loss == "cross_entropy":
            seq += [Sigmoid()]
        self.seq = Sequential(*seq)

    def dragan_penalty(self, real_data, device="cpu", pac=10, lambda_=10):
        # real_data = torch.from_numpy(real_data).to(device)
        alpha = (
            torch.rand(real_data.shape[0], 1, device=device)
            .squeeze()
            .expand(real_data.shape[0])
        )
        delta = torch.normal(
            mean=0.0, std=float(pac), size=real_data.shape, device=device
        )  # 0.5 * real_data.std() * torch.rand(real_data.shape)
        x_hat = Variable(
            (alpha * real_data.T + (1 - alpha) * (real_data + delta).T).T,
            requires_grad=True,
        )

        pred_hat = self(x_hat.float())

        gradients = torch.autograd.grad(
            outputs=pred_hat,
            inputs=x_hat,
            grad_outputs=torch.ones(pred_hat.size(), device=device),
            create_graph=True,
            retain_graph=True,
            only_inputs=True,
        )[0]
        dragan_penalty = lambda_ * ((gradients.norm(2, dim=1) - 1) ** 2).mean()

        return dragan_penalty

    def forward(self, input):
        assert input.size()[0] % self.pac == 0
        return self.seq(input.view(-1, self.pacdim))


class Residual(Module):
    def __init__(self, i, o):
        super(Residual, self).__init__()
        self.fc = Linear(i, o)
        self.bn = BatchNorm1d(o)
        self.relu = ReLU()

    def forward(self, input):
        out = self.fc(input)
        out = self.bn(out)
        out = self.relu(out)
        return torch.cat([out, input], dim=1)


class Generator(Module):
    def __init__(self, embedding_dim, generator_dim, data_dim):
        super(Generator, self).__init__()
        dim = embedding_dim
        seq = []
        for item in list(generator_dim):
            seq += [Residual(dim, item)]
            dim += item
        seq.append(Linear(dim, data_dim))
        self.seq = Sequential(*seq)

    def forward(self, input):
        data = self.seq(input)
        return data


class PATECTGAN(CTGANSynthesizer, Synthesizer):
    def __init__(
        self,
        embedding_dim=128,
        generator_dim=(256, 256),
        discriminator_dim=(256, 256),
        generator_lr=2e-4,
        generator_decay=1e-6,
        discriminator_lr=2e-4,
        discriminator_decay=1e-6,
        batch_size=500,
        discriminator_steps=1,
        verbose=True,
        epochs=300,
        pac=1,
        cuda=True,
        epsilon=1,
        binary=False,
        regularization=None,
        loss="cross_entropy",
        teacher_iters=5,
        student_iters=5,
        sample_per_teacher=1000,
        delta=None,
        noise_multiplier=1e-3,
        moments_order=100,
    ):

        assert batch_size % 2 == 0

        self._embedding_dim = embedding_dim
        self._generator_dim = generator_dim
        self._discriminator_dim = discriminator_dim

        self._generator_lr = generator_lr
        self._generator_decay = generator_decay
        self._discriminator_lr = discriminator_lr
        self._discriminator_decay = discriminator_decay

        self._batch_size = batch_size
        self._discriminator_steps = discriminator_steps
        self._verbose = verbose
        self._epochs = epochs
        self.pac = pac
        self.epsilon = epsilon
        self.verbose = verbose
        self.loss = loss

        # PATE params
        self.regularization = regularization if self.loss != "wasserstein" else "dragan"
        self.teacher_iters = teacher_iters
        self.student_iters = student_iters
        self.pd_cols = None
        self.pd_index = None
        self.binary = binary
        self.sample_per_teacher = sample_per_teacher
        self.noise_multiplier = noise_multiplier
        self.moments_order = moments_order
        self.delta = delta

        if not cuda or not torch.cuda.is_available():
            device = "cpu"
        elif isinstance(cuda, str):
            device = cuda
        else:
            device = "cuda"

        self._device = torch.device(device)

    def train(
        self,
        data,
        categorical_columns=None,
        ordinal_columns=None,
        update_epsilon=None,
        transformer=None,
        continuous_columns=None, 
        preprocessor_eps=0.0,
        nullable=False
    ):
        if update_epsilon:
            self.epsilon = update_epsilon

        sample_per_teacher = (
            self.sample_per_teacher if self.sample_per_teacher < len(data) else 1000
        )
        self.num_teachers = int(len(data) / sample_per_teacher) + 1

        train_data = self._get_train_data(
            data,
            style='gan',
            transformer=transformer,
            categorical_columns=categorical_columns, 
            ordinal_columns=ordinal_columns, 
            continuous_columns=continuous_columns, 
            nullable=nullable,
            preprocessor_eps=preprocessor_eps
        )

        train_data = np.array([
            [float(x) if x is not None else 0.0 for x in row] for row in train_data
        ])

        data_partitions = np.array_split(train_data, self.num_teachers)

        data_dim = self._transformer.output_width


        self.cond_generator = DataSampler(
            train_data,
            self._transformer.transformers
        )

        cached_probs = self.cond_generator.discrete_column_category_prob

        cond_generator = [
            DataSampler(
                d,
                self._transformer.transformers,
                discrete_column_category_prob=cached_probs,
            )
            for d in data_partitions
        ]

        self._generator = Generator(
            self._embedding_dim + self.cond_generator.dim_cond_vec(),
            self._generator_dim,
            data_dim,
        ).to(self._device)

        discriminator = Discriminator(
            data_dim + self.cond_generator.dim_cond_vec(),
            self._discriminator_dim,
            self.loss,
            self.pac,
        ).to(self._device)

        student_disc = discriminator
        student_disc.apply(weights_init)

        teacher_disc = [discriminator for i in range(self.num_teachers)]
        for i in range(self.num_teachers):
            teacher_disc[i].apply(weights_init)

        optimizerG = optim.Adam(
            self._generator.parameters(),
            lr=self._generator_lr,
            betas=(0.5, 0.9),
            weight_decay=self._generator_decay,
        )

        optimizer_s = optim.Adam(student_disc.parameters(), lr=2e-4, betas=(0.5, 0.9))
        optimizer_t = [
            optim.Adam(
                teacher_disc[i].parameters(),
                lr=self._discriminator_lr,
                betas=(0.5, 0.9),
                weight_decay=self._discriminator_decay,
            )
            for i in range(self.num_teachers)
        ]

        noise_multiplier = self.noise_multiplier
        alphas = torch.tensor(
            [0.0 for i in range(self.moments_order)], device=self._device
        )
        l_list = 1 + torch.tensor(range(self.moments_order), device=self._device)
        eps = torch.zeros(1)

        mean = torch.zeros(self._batch_size, self._embedding_dim, device=self._device)
        std = mean + 1

        real_label = 1
        fake_label = 0

        criterion = nn.BCELoss() if (self.loss == "cross_entropy") else self.w_loss

        if self.verbose:
            print(
                "using loss {} and regularization {}".format(
                    self.loss, self.regularization
                )
            )

        iteration = 0

        if self.delta is None:
            self.delta = 1 / (train_data.shape[0] * np.sqrt(train_data.shape[0]))

        while eps.item() < self.epsilon:
            iteration += 1

            eps = min((alphas - math.log(self.delta)) / l_list)

            if eps.item() > self.epsilon:
                if iteration == 1:
                    raise ValueError(
                        "Inputted epsilon parameter is too small to"
                        + " create a private dataset. Try increasing epsilon and rerunning."
                    )
                break

            # train teacher discriminators
            for t_2 in range(self.teacher_iters):
                for i in range(self.num_teachers):
                    partition_data = data_partitions[i]
                    data_sampler = DataSampler(
                        partition_data,
                        self._transformer.transformers,
                        discrete_column_category_prob=cached_probs,
                    )
                    fakez = torch.normal(mean, std=std).to(self._device)

                    condvec = cond_generator[i].sample_condvec(self._batch_size)

                    if condvec is None:
                        c1, m1, col, opt = None, None, None, None
                        real = data_sampler.sample_data(self._batch_size, col, opt)
                    else:
                        c1, m1, col, opt = condvec
                        c1 = torch.from_numpy(c1).to(self._device)
                        m1 = torch.from_numpy(m1).to(self._device)
                        fakez = torch.cat([fakez, c1], dim=1)
                        perm = np.arange(self._batch_size)
                        np.random.shuffle(perm)
                        real = data_sampler.sample_data(
                            self._batch_size, col[perm], opt[perm]
                        )
                        c2 = c1[perm]

                    fake = self._generator(fakez)
                    fakeact = self._apply_activate(fake)

                    real = torch.from_numpy(real.astype("float32")).to(self._device)

                    if c1 is not None:
                        fake_cat = torch.cat([fakeact, c1], dim=1)
                        real_cat = torch.cat([real, c2], dim=1)
                    else:
                        real_cat = real
                        fake_cat = fake

                    optimizer_t[i].zero_grad()

                    y_all = torch.cat(
                        [teacher_disc[i](fake_cat), teacher_disc[i](real_cat)]
                    )
                    label_fake = torch.full(
                        (int(self._batch_size / self.pac), 1),
                        fake_label,
                        dtype=torch.float,
                        device=self._device,
                    )
                    label_true = torch.full(
                        (int(self._batch_size / self.pac), 1),
                        real_label,
                        dtype=torch.float,
                        device=self._device,
                    )
                    labels = torch.cat([label_fake, label_true])

                    error_d = criterion(y_all.squeeze(), labels.squeeze())
                    error_d.backward()

                    if self.regularization == "dragan":
                        pen = teacher_disc[i].dragan_penalty(
                            real_cat, device=self._device
                        )
                        pen.backward(retain_graph=True)

                    optimizer_t[i].step()
            ###
            # train student discriminator
            for t_3 in range(self.student_iters):
                data_sampler = DataSampler(
                    train_data,
                    self._transformer.transformers,
                    discrete_column_category_prob=cached_probs,
                )
                fakez = torch.normal(mean=mean, std=std)

                condvec = self.cond_generator.sample_condvec(self._batch_size)

                if condvec is None:
                    c1, m1, col, opt = None, None, None, None
                    real = data_sampler.sample_data(self._batch_size, col, opt)
                else:
                    c1, m1, col, opt = condvec
                    c1 = torch.from_numpy(c1).to(self._device)
                    m1 = torch.from_numpy(m1).to(self._device)
                    fakez = torch.cat([fakez, c1], dim=1)

                    perm = np.arange(self._batch_size)
                    np.random.shuffle(perm)
                    real = data_sampler.sample_data(
                        self._batch_size, col[perm], opt[perm]
                    )
                    c2 = c1[perm]

                fake = self._generator(fakez)
                fakeact = self._apply_activate(fake)

                if c1 is not None:
                    fake_cat = torch.cat([fakeact, c1], dim=1)
                else:
                    fake_cat = fakeact

                fake_data = fake_cat

                ###
                predictions, votes = pate(
                    fake_data, teacher_disc, noise_multiplier, device=self._device
                )

                output = student_disc(fake_data.detach())

                # update moments accountant
                alphas = alphas + moments_acc(
                    self.num_teachers,
                    votes,
                    noise_multiplier,
                    l_list,
                    device=self._device,
                )

                loss_s = criterion(
                    output.squeeze(), predictions.float().to(self._device).squeeze()
                )

                optimizer_s.zero_grad()
                loss_s.backward()

                if self.regularization == "dragan":
                    vals = torch.cat([predictions, fake_data], axis=1)
                    ordered = vals[vals[:, 0].sort()[1]]
                    data_list = torch.split(
                        ordered, predictions.shape[0] - int(predictions.sum().item())
                    )
                    synth_cat = torch.cat(data_list[1:], axis=0)[:, 1:]
                    pen = student_disc.dragan_penalty(synth_cat, device=self._device)
                    pen.backward(retain_graph=True)

                optimizer_s.step()

                # print ('iterator {i}, student discriminator loss is {j}'.format(i=t_3, j=loss_s))

            # train generator
            fakez = torch.normal(mean=mean, std=std)
            condvec = self.cond_generator.sample_condvec(self._batch_size)

            if condvec is None:
                c1, m1, col, opt = None, None, None, None
            else:
                c1, m1, col, opt = condvec
                c1 = torch.from_numpy(c1).to(self._device)
                m1 = torch.from_numpy(m1).to(self._device)
                fakez = torch.cat([fakez, c1], dim=1)

            fake = self._generator(fakez)
            fakeact = self._apply_activate(fake)

            if c1 is not None:
                y_fake = student_disc(torch.cat([fakeact, c1], dim=1))
            else:
                y_fake = student_disc(fakeact)

            if condvec is None:
                cross_entropy = 0
            else:
                cross_entropy = self._cond_loss(fake, c1, m1)

            if self.loss == "cross_entropy":
                label_g = torch.full(
                    (int(self._batch_size / self.pac), 1),
                    real_label,
                    dtype=torch.float,
                    device=self._device,
                )
                loss_g = criterion(y_fake.squeeze(), label_g.float().squeeze())
                loss_g = loss_g + cross_entropy
            else:
                loss_g = -torch.mean(y_fake) + cross_entropy

            optimizerG.zero_grad()
            loss_g.backward()
            optimizerG.step()

            if self.verbose:
                print(
                    "eps: {:f} \t G: {:f} \t D: {:f}".format(
                        eps, loss_g.detach().cpu(), loss_s.detach().cpu()
                    )
                )

    def w_loss(self, output, labels):
        vals = torch.cat([labels[None, :], output[None, :]], axis=1)
        ordered = vals[vals[:, 0].sort()[1]]
        data_list = torch.split(ordered, labels.shape[0] - int(labels.sum().item()), dim=1)
        fake_score = data_list[0][:, 1]
        true_score = torch.cat(data_list[1:], axis=0)[:, 1]
        w_loss = -(torch.mean(true_score) - torch.mean(fake_score))
        return w_loss

    def generate(self, n, condition_column=None, condition_value=None):
        """
        TODO: Add condition_column support 
        """
        self._generator.eval()

        # output_info = self._transformer.output_info
        steps = n // self._batch_size + 1
        data = []
        for i in range(steps):
            mean = torch.zeros(self._batch_size, self._embedding_dim)
            std = mean + 1
            fakez = torch.normal(mean=mean, std=std).to(self._device)

            condvec = self.cond_generator.sample_original_condvec(self._batch_size)

            if condvec is None:
                pass
            else:
                c1 = condvec
                c1 = torch.from_numpy(c1).to(self._device)
                fakez = torch.cat([fakez, c1], dim=1)

            fake = self._generator(fakez)
            fakeact = self._apply_activate(fake)
            data.append(fakeact.detach().cpu().numpy())

        data = np.concatenate(data, axis=0)
        data = data[:n]

        return self._transformer.inverse_transform(data)

    def fit(self, data, *ignore, transformer=None, categorical_columns=[], ordinal_columns=[], continuous_columns=[], preprocessor_eps=0.0, nullable=False):
        self.train(data, transformer=transformer, categorical_columns=categorical_columns, ordinal_columns=ordinal_columns, continuous_columns=continuous_columns, preprocessor_eps=preprocessor_eps, nullable=nullable)

    def sample(self, n_samples):
        return self.generate(n_samples)



import math
import numpy as np
import pandas as pd

import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader, TensorDataset

from snsynth.base import Synthesizer

from ._generator import Generator
from ._discriminator import Discriminator

from .privacy_utils import weights_init, pate, moments_acc


class PATEGAN(Synthesizer):
    def __init__(
        self,
        epsilon,
        delta=None,
        binary=False,
        latent_dim=64,
        batch_size=64,
        teacher_iters=5,
        student_iters=5,
    ):
        self.epsilon = epsilon
        self.delta = delta
        self.binary = binary
        self.latent_dim = latent_dim
        self.batch_size = batch_size
        self.teacher_iters = teacher_iters
        self.student_iters = student_iters

        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

        self.pd_cols = None
        self.pd_index = None

    def train(
        self,
        data,
        categorical_columns=None,
        ordinal_columns=None,
        update_epsilon=None,
        transformer=None,
        continuous_columns=None,
        preprocessor_eps=0.0,
        nullable=False,
    ):
        if update_epsilon:
            self.epsilon = update_epsilon

        train_data = self._get_train_data(
            data,
            style='gan',
            transformer=transformer,
            categorical_columns=categorical_columns, 
            ordinal_columns=ordinal_columns, 
            continuous_columns=continuous_columns, 
            nullable=nullable,
            preprocessor_eps=preprocessor_eps
        )

        data = np.array(train_data)

        if isinstance(data, pd.DataFrame):
            for col in data.columns:
                data[col] = pd.to_numeric(data[col], errors="ignore")
            self.pd_cols = data.columns
            self.pd_index = data.index
            data = data.to_numpy()
        elif not isinstance(data, np.ndarray):
            raise ValueError("Data must be a numpy array or pandas dataframe")

        data_dim = data.shape[1]

        self.num_teachers = int(len(data) / 1000)

        data_partitions = np.array_split(data, self.num_teachers)
        tensor_partitions = [
            TensorDataset(torch.from_numpy(data.astype("double")).to(self.device))
            for data in data_partitions
        ]

        loader = []
        for teacher_id in range(self.num_teachers):
            loader.append(
                DataLoader(
                    tensor_partitions[teacher_id],
                    batch_size=self.batch_size,
                    shuffle=True,
                )
            )

        self.generator = (
            Generator(self.latent_dim, data_dim, binary=self.binary)
            .double()
            .to(self.device)
        )
        self.generator.apply(weights_init)

        student_disc = Discriminator(data_dim).double().to(self.device)
        student_disc.apply(weights_init)

        teacher_disc = [
            Discriminator(data_dim).double().to(self.device)
            for i in range(self.num_teachers)
        ]
        for i in range(self.num_teachers):
            teacher_disc[i].apply(weights_init)

        optimizer_g = optim.Adam(self.generator.parameters(), lr=1e-4)
        optimizer_s = optim.Adam(student_disc.parameters(), lr=1e-4)
        optimizer_t = [
            optim.Adam(teacher_disc[i].parameters(), lr=1e-4)
            for i in range(self.num_teachers)
        ]

        criterion = nn.BCELoss()

        noise_multiplier = 1e-3
        alphas = torch.tensor([0.0 for i in range(100)])
        l_list = 1 + torch.tensor(range(100))
        eps = torch.zeros(1)

        if self.delta is None:
            self.delta = 1 / (data.shape[0] * np.sqrt(data.shape[0]))

        iteration = 0
        while eps.item() < self.epsilon:
            iteration += 1

            eps = min((alphas - math.log(self.delta)) / l_list)

            if eps.item() > self.epsilon:
                if iteration == 1:
                    raise ValueError(
                        "Inputted epsilon parameter is too small to"
                        + " create a private dataset. Try increasing epsilon and rerunning."
                    )
                break

            # train teacher discriminators
            for t_2 in range(self.teacher_iters):
                for i in range(self.num_teachers):
                    real_data = None
                    for j, data in enumerate(loader[i], 0):
                        real_data = data[0].to(self.device)
                        break

                    optimizer_t[i].zero_grad()

                    # train with real data
                    label_real = torch.full(
                        (real_data.shape[0],), 1, dtype=torch.float, device=self.device
                    )
                    output = teacher_disc[i](real_data)
                    loss_t_real = criterion(output.squeeze(), label_real.double())
                    loss_t_real.backward()

                    # train with fake data
                    noise = torch.rand(
                        self.batch_size, self.latent_dim, device=self.device
                    )
                    label_fake = torch.full(
                        (self.batch_size,), 0, dtype=torch.float, device=self.device
                    )
                    fake_data = self.generator(noise.double())
                    output = teacher_disc[i](fake_data)
                    loss_t_fake = criterion(output.squeeze(), label_fake.double())
                    loss_t_fake.backward()
                    optimizer_t[i].step()

            # train student discriminator
            for t_3 in range(self.student_iters):
                noise = torch.rand(self.batch_size, self.latent_dim, device=self.device)
                fake_data = self.generator(noise.double())
                predictions, votes = pate(fake_data, teacher_disc, noise_multiplier)
                output = student_disc(fake_data.detach())

                # update moments accountant
                alphas = alphas + moments_acc(
                    self.num_teachers, votes, noise_multiplier, l_list
                )

                loss_s = criterion(
                    output.squeeze(), predictions.to(self.device).squeeze()
                )
                optimizer_s.zero_grad()
                loss_s.backward()
                optimizer_s.step()

            # train generator
            label_g = torch.full(
                (self.batch_size,), 1, dtype=torch.float, device=self.device
            )
            noise = torch.rand(self.batch_size, self.latent_dim, device=self.device)
            gen_data = self.generator(noise.double())
            output_g = student_disc(gen_data)
            loss_g = criterion(output_g.squeeze(), label_g.double())
            optimizer_g.zero_grad()
            loss_g.backward()
            optimizer_g.step()

    def generate(self, n):
        steps = n // self.batch_size + 1
        data = []
        for i in range(steps):
            noise = torch.randn(self.batch_size, self.latent_dim, device=self.device)
            noise = noise.view(-1, self.latent_dim)

            fake_data = self.generator(noise.double())
            data.append(fake_data.detach().cpu().numpy())

        data = np.concatenate(data, axis=0)
        data = data[:n]

        return self._transformer.inverse_transform(data)

    def fit(self, data, *ignore, transformer=None, categorical_columns=[], ordinal_columns=[], continuous_columns=[], preprocessor_eps=0.0, nullable=False):
        self.train(data, transformer=transformer, categorical_columns=categorical_columns, ordinal_columns=ordinal_columns, continuous_columns=continuous_columns, preprocessor_eps=preprocessor_eps, nullable=nullable)

    def sample(self, n_samples):
        return self.generate(n_samples)



from .base import ColumnTransformer

class IdentityTransformer(ColumnTransformer):
    """Dummy transformer that passes through values.  Useful for testing.
    """
    def __init__(self):
        super().__init__()
    @property
    def cardinality(self):
        return [None]
    def _fit(self, val):
        pass
    def _clear_fit(self):
        self._fit_complete = True
        self.output_width = 1
    def _transform(self, val):
        return val
    def _inverse_transform(self, val):
        return val



import numpy as np
import pandas as pd

try:
    from mbi import FactoredInference, Dataset, Domain
except ImportError:
    print("Please install mbi with:\n   pip install git+https://github.com/ryan112358/private-pgm.git")

from scipy import sparse
from disjoint_set import DisjointSet
import networkx as nx
import itertools
from scipy.special import logsumexp
from snsynth.mst.cdp2adp import cdp_rho
from snsynth.base import Synthesizer

"""
Wrapper for MST synthesizer from Private PGM:
https://github.com/ryan112358/private-pgm/tree/e9ea5fcac62e2c5b92ae97f7afe2648c04432564

This is a generalization of the winning mechanism from the
2018 NIST Differential Privacy Synthetic Data Competition.

Unlike the original implementation, this one can work for any discrete dataset,
and does not rely on public provisional data for measurement selection.
"""


class MSTSynthesizer(Synthesizer):
    """Maximum Spanning Tree synthesizer, uses Private PGM.

    :param epsilon: privacy budget for the synthesizer
    :type epsilon: float
    :param delta: privacy parameter.  Should be small, in the range of 1/(n * sqrt(n))
    :type delta: float
    :param verbose: print diagnostic information during processing
    :type verbose: bool

    Reuses code and modifies it lightly from
    https://github.com/ryan112358/private-pgm/blob/master/mechanisms/mst.py. Awesome work McKenna et. al!
    """

    def __init__(self,
                 epsilon=0.1,
                 delta=1e-9,
                 *ignore,
                 verbose=False
                 ):

        self.epsilon = epsilon
        self.delta = delta
        self.verbose = verbose

        self.synthesizer = None
        self.num_rows = None

    def fit(
        self, 
        data, 
        *ignore,
        transformer=None,
        categorical_columns=[], 
        ordinal_columns=[],
        continuous_columns=[],
        preprocessor_eps=0.0,
        nullable=False,
        prng=np.random
        ):

        train_data = self._get_train_data(
            data,
            style='cube',
            transformer=transformer,
            categorical_columns=categorical_columns, 
            ordinal_columns=ordinal_columns, 
            continuous_columns=continuous_columns, 
            nullable=nullable,
            preprocessor_eps=preprocessor_eps
        )

        if self._transformer is None:
            raise ValueError("We weren't able to fit a transformer to the data. Please check your data and try again.")

        cards = self._transformer.cardinality
        if any (c is None for c in cards):
            raise ValueError("The transformer appears to have some continuous columns. Please provide only categorical or ordinal.")

        dimensionality = np.prod(cards)
        if self.verbose:
            print(f"Fitting with {dimensionality} dimensions")

        colnames = ["col" + str(i) for i in range(self._transformer.output_width)]
        cards = self._transformer.cardinality

        if len(cards) != len(colnames):
            raise ValueError("Cardinality and column names must be the same length.")

        domain = Domain(colnames, cards)
        self.num_rows = len(data)

        data = pd.DataFrame(train_data, columns=colnames)
        data = Dataset(df=data, domain=domain)

        self.MST(data, self.epsilon, self.delta)

    def sample(self, samples=None):
        if samples is None:
            samples = self.num_rows
        data = self.synthesizer.synthetic_data(rows=samples)
        decompressed = self.undo_compress_fn(data)
        data_iter = [tuple([c for c in t[1:]]) for t in decompressed.df.itertuples()]
        return self._transformer.inverse_transform(data_iter)

    def MST(self, data, epsilon, delta):
        rho = cdp_rho(epsilon, delta)
        sigma = np.sqrt(3/(2*rho))
        cliques = [(col,) for col in data.domain]
        if self.verbose:
            print("Getting cliques")
        log1 = self.measure(data, cliques, sigma)
        data, log1, undo_compress_fn = self.compress_domain(data, log1)

        # Here's the decompress function
        self.undo_compress_fn = undo_compress_fn

        cliques = self.select(data, rho/3.0, log1)
        log2 = self.measure(data, cliques, sigma)
        engine = FactoredInference(data.domain, iters=5000)
        if self.verbose:
            print("Estimating marginals")
        est = engine.estimate(log1+log2)

        # Here's the synthesizer

        self.synthesizer = est

    def measure(self, data, cliques, sigma, weights=None):
        if weights is None:
            weights = np.ones(len(cliques))
        weights = np.array(weights) / np.linalg.norm(weights)
        measurements = []
        for proj, wgt in zip(cliques, weights):
            x = data.project(proj).datavector()
            y = x + np.random.normal(loc=0, scale=sigma/wgt, size=x.size)
            Q = sparse.eye(x.size)
            measurements.append((Q, y, sigma/wgt, proj))
        return measurements

    def compress_domain(self, data, measurements):
        supports = {}
        new_measurements = []
        for Q, y, sigma, proj in measurements:
            col = proj[0]
            sup = y >= 3*sigma
            supports[col] = sup
            if supports[col].sum() == y.size:
                new_measurements.append((Q, y, sigma, proj))
            else:  # need to re-express measurement over the new domain
                y2 = np.append(y[sup], y[~sup].sum())
                I2 = np.ones(y2.size)
                I2[-1] = 1.0 / np.sqrt(y.size - y2.size + 1.0)
                y2[-1] /= np.sqrt(y.size - y2.size + 1.0)
                I2 = sparse.diags(I2)
                new_measurements.append((I2, y2, sigma, proj))

        undo_compress_fn = lambda data: self.reverse_data(data, supports)  # noqa: E731

        return self.transform_data(data, supports), new_measurements, undo_compress_fn

    def exponential_mechanism(self, q, eps, sensitivity, prng=np.random, monotonic=False):
        coef = 1.0 if monotonic else 0.5
        scores = coef*eps/sensitivity*q
        probas = np.exp(scores - logsumexp(scores))
        return prng.choice(q.size, p=probas)

    def select(self, data, rho, measurement_log, cliques=[]):
        engine = FactoredInference(data.domain, iters=1000)
        est = engine.estimate(measurement_log)

        weights = {}
        candidates = list(itertools.combinations(data.domain.attrs, 2))
        for a, b in candidates:
            xhat = est.project([a, b]).datavector()
            x = data.project([a, b]).datavector()
            weights[a, b] = np.linalg.norm(x - xhat, 1)

        T = nx.Graph()
        T.add_nodes_from(data.domain.attrs)
        ds = DisjointSet()

        for e in cliques:
            T.add_edge(*e)
            ds.union(*e)

        r = len(list(nx.connected_components(T)))
        epsilon = np.sqrt(8*rho/(r-1))
        for i in range(r-1):
            candidates = [e for e in candidates if not ds.connected(*e)]
            wgts = np.array([weights[e] for e in candidates])
            idx = self.exponential_mechanism(wgts, epsilon, sensitivity=1.0)
            e = candidates[idx]
            T.add_edge(*e)
            ds.union(*e)

        return list(T.edges)

    def transform_data(self, data, supports):
        df = data.df.copy()
        newdom = {}
        for col in data.domain:
            support = supports[col]
            size = support.sum()
            newdom[col] = int(size)
            if size < support.size:
                newdom[col] += 1
            mapping = {}
            idx = 0
            for i in range(support.size):
                mapping[i] = size
                if support[i]:
                    mapping[i] = idx
                    idx += 1
            assert idx == size
            df[col] = df[col].map(mapping)
        newdom = Domain.fromdict(newdom)
        return Dataset(df, newdom)

    def reverse_data(self, data, supports):
        df = data.df.copy()
        newdom = {}
        for col in data.domain:
            support = supports[col]
            mx = support.sum()
            newdom[col] = int(support.size)
            idx, extra = np.where(support)[0], np.where(~support)[0]
            mask = df[col] == mx
            if extra.size == 0:
                pass
            else:
                df.loc[mask, col] = np.random.choice(extra, mask.sum())
            df.loc[~mask, col] = idx[df.loc[~mask, col]]
        newdom = Domain.fromdict(newdom)
        return Dataset(df, newdom)



import warnings
import numpy as np


class DataSampler(object):
    """DataSampler samples the conditional vector and corresponding data for CTGAN.

    :param data(np.ndarray): The data to be conditionally sampled.
    :param output_info(list): The output information of the model.
    :param discrete_column_category_prob(list): The category probabilities for each discrete column.
        Use this to pass in cached noisy probabilities.  Data must match the schema of the original data.
    """

    def __init__(
            self,
            data,
            transformers,
            *ignore,
            discrete_column_category_prob=None,
            **kwargs):
        self._data = data
        self._transformers = transformers

        self._per_column_scale = None
        if discrete_column_category_prob is not None:
            warnings.warn("Discrete column category prob is deprecated and will be removed in a future version.")
        self._discrete_column_category_prob = None

        n_discrete_columns = sum(
            [1 for t in self._transformers if t.is_categorical])

        self._discrete_column_matrix_st = np.zeros(
            n_discrete_columns, dtype="int32")

        # Store the row id for each category in each discrete column.
        # For example _rid_by_cat_cols[a][b] is a list of all rows with the
        # a-th discrete column equal value b.
        self._rid_by_cat_cols = []

        # Compute _rid_by_cat_cols
        st = 0
        for t in self._transformers:
            if t.is_categorical:
                ed = st + t.output_width

                rid_by_cat = []
                for j in range(t.output_width):
                    rid_by_cat.append(np.nonzero(data[:, st + j])[0])
                self._rid_by_cat_cols.append(rid_by_cat)
                st = ed
            else:
                st += t.output_width
        assert st == data.shape[1]

        # Prepare an interval matrix for efficiently sample conditional vector
        max_category = max(
            [t.output_width for t in self._transformers
             if t.is_categorical], default=0)

        self._discrete_column_cond_st = np.zeros(n_discrete_columns, dtype='int32')
        self._discrete_column_n_category = np.zeros(
            n_discrete_columns, dtype='int32')
        self._discrete_column_category_prob = np.zeros(
            (n_discrete_columns, max_category))
        self._n_discrete_columns = n_discrete_columns
        self._n_categories = sum(
            [t.output_width for t in self._transformers
             if t.is_categorical])

        eps_tot = 0.0
        st = 0
        current_id = 0
        current_cond_st = 0
        for t in self._transformers:
            if t.is_categorical:
                ed = st + t.output_width
                category_freq = np.sum(data[:, st:ed], axis=0)
                category_freq = [1 if v < 1 else v for v in category_freq]
                category_freq = np.array(category_freq, dtype='float64')
                category_prob = category_freq / np.sum(category_freq)
                self._discrete_column_category_prob[current_id, :t.output_width] = (
                    category_prob)
                self._discrete_column_cond_st[current_id] = current_cond_st
                self._discrete_column_n_category[current_id] = t.output_width
                current_cond_st += t.output_width
                current_id += 1
                st = ed
            else:
                st += t.output_width

        if discrete_column_category_prob is not None:
            assert len(discrete_column_category_prob) == n_discrete_columns
            for i in range(n_discrete_columns):
                self._discrete_column_category_prob[i, :] = discrete_column_category_prob[i]
            self.total_spent = 0.0  # don't have to pay for cached noise

    @property
    def discrete_column_category_prob(self):
        return self._discrete_column_category_prob

    def _random_choice_prob_index(self, discrete_column_id):
        probs = self._discrete_column_category_prob[discrete_column_id]
        r = np.expand_dims(np.random.rand(probs.shape[0]), axis=1)
        return (probs.cumsum(axis=1) > r).argmax(axis=1)

    def sample_condvec(self, batch):
        """Generate the conditional vector for training.

        Returns:
            cond (batch x #categories):
                The conditional vector.
            mask (batch x #discrete columns):
                A one-hot vector indicating the selected discrete column.
            discrete column id (batch):
                Integer representation of mask.
            category_id_in_col (batch):
                Selected category in the selected discrete column.
        """
        if self._n_discrete_columns == 0:
            return None

        discrete_column_id = np.random.choice(
            np.arange(self._n_discrete_columns), batch)

        cond = np.zeros((batch, self._n_categories), dtype='float32')
        mask = np.zeros((batch, self._n_discrete_columns), dtype='float32')
        mask[np.arange(batch), discrete_column_id] = 1
        category_id_in_col = self._random_choice_prob_index(discrete_column_id)
        category_id = (self._discrete_column_cond_st[discrete_column_id]
                       + category_id_in_col)
        cond[np.arange(batch), category_id] = 1

        return cond, mask, discrete_column_id, category_id_in_col

    def sample_original_condvec(self, batch):
        """Generate the conditional vector for generation use original frequency."""
        if self._n_discrete_columns == 0:
            return None

        cond = np.zeros((batch, self._n_categories), dtype='float32')

        for i in range(batch):
            row_idx = np.random.randint(0, len(self._data))
            col_idx = np.random.randint(0, self._n_discrete_columns)
            matrix_st = self._discrete_column_matrix_st[col_idx]
            matrix_ed = matrix_st + self._discrete_column_n_category[col_idx]
            pick = np.argmax(self._data[row_idx, matrix_st:matrix_ed])
            cond[i, pick + self._discrete_column_cond_st[col_idx]] = 1

        return cond

    def sample_data(self, n, col, opt):
        """Sample data from original training data satisfying the sampled conditional vector.

        Returns:
            n rows of matrix data.
        """
        if col is None:
            idx = np.random.randint(len(self._data), size=n)
            return self._data[idx]

        idx = []
        for c, o in zip(col, opt):
            if len(self._rid_by_cat_cols[c][o]) == 0:
                # if teacher splits result in zero probability for a category value
                idx.append(np.random.randint(len(self._data)))
            else:
                idx.append(np.random.choice(self._rid_by_cat_cols[c][o]))

        return self._data[idx]

    def dim_cond_vec(self):
        return self._n_categories

    def generate_cond_from_condition_column_info(self, condition_info, batch):
        vec = np.zeros((batch, self._n_categories), dtype='float32')
        id = self._discrete_column_matrix_st[condition_info["discrete_column_id"]
                                             ] + condition_info["value_id"]
        vec[:, id] = 1
        return vec



from .base import ColumnTransformer
from .definitions import ColumnType
from faker import Faker


class AnonymizationTransformer(ColumnTransformer):
    """
    Transformer that can be used to anonymize personally identifiable information (PII) or other values.
    By default, the existing values are discarded during transformation and not used by a synthesizer.
    During inverse transformation new values will be generated according to the specified ``fake``.

    If ``fake_inbound`` is true, the new values will be injected during transformation and passed through on inverse.
    This might be useful for e.g. operation in a ChainTransformer.

    Beware that the provided ``fake`` is called once to verify that the provided (keyword) arguments are valid.

    :param fake: Text reference to Faker method (e.g. 'email') or custom callable
    :type fake: str or callable, required
    :param args: Arguments for the method
    :type args: args, optional
    :param faker_setup: Dictionary with keyword arguments for Faker initialization e.g. {'locale': 'de_DE'}
    :type faker_setup: dict, optional
    :param fake_inbound: Defaults to False.
    :type fake_inbound: bool, optional
    :param kwargs: Keyword arguments for the method
    :type kwargs: kwargs, optional
    """

    def __init__(self, fake, *args, faker_setup=None, fake_inbound=False, **kwargs):
        self.fake_inbound = fake_inbound
        super().__init__()

        if isinstance(fake, str):  # assume this references a Faker builtin
            fake = self._get_faker_builtin(fake, faker_setup, *args, **kwargs)

        self.fake = fake
        self.args = args
        self.kwargs = kwargs

        # verify that the provided arguments are valid
        try:
            self._generate_fake_data()
        except TypeError as e:
            raise ValueError(
                f"Provided arguments {args} and {kwargs} are invalid for `fake` {fake}"
            ) from e

    def _get_faker_builtin(self, fake, faker_setup, *args, **kwargs):
        """
        Creates a Faker instance and verifies that the given method is available.

        :param fake: Text reference to Faker method
        :type fake: str, required
        :param faker_setup: Dictionary with keyword arguments for initializing Faker e.g. {'locale': 'de_DE'}
        :type faker_setup: dict, optional
        :param args: Arguments for the method
        :type args: args, optional
        :param kwargs: Keyword arguments for the method
        :type kwargs: kwargs, optional
        :return: Actual Faker method
        :rtype: callable
        """
        # initialize Faker with provided setup or default
        if isinstance(faker_setup, dict):
            try:
                self.faker = Faker(**faker_setup)
            except Exception as e:
                raise ValueError(
                    f"Provided `faker_setup` {faker_setup} is invalid"
                ) from e
        else:
            self.faker = Faker()

        # verify that the provided fake is available
        try:
            fake_builtin = getattr(self.faker, fake)
        except AttributeError as e:
            raise ValueError(f"Provided `fake` {fake} is not available in Faker") from e

        return fake_builtin

    @property
    def output_type(self):
        return ColumnType.UNBOUNDED

    @property
    def cardinality(self):
        return [None]

    def _fit(self, _):
        pass

    def _clear_fit(self):
        self._fit_complete = True
        self.output_width = 1 if self.fake_inbound else 0

    def _generate_fake_data(self):
        return self.fake(*self.args, **self.kwargs)

    def _transform(self, _):
        if self.fake_inbound:
            return self._generate_fake_data()
        else:
            return None

    def _inverse_transform(self, val):
        if self.fake_inbound:
            return val
        else:
            return self._generate_fake_data()

    def transform(self, data, idx=None):
        if idx is None:
            return [self._transform(val) for val in data]
        else:
            return [row[:idx] + row[idx + 1 :] for row in data]

    def inverse_transform(self, data, idx=None):
        if idx is None:
            return [self._inverse_transform(val) for val in data]
        else:
            return [
                row[:idx] + (self._inverse_transform(None),) + row[idx:] for row in data
            ]



from snsynth.transform.definitions import ColumnType
from .base import CachingColumnTransformer
from opendp.mod import enable_features, binary_search_param
from opendp.transformations import make_sized_bounded_mean, make_sized_bounded_variance, make_clamp, make_bounded_resize
from opendp.measurements import make_base_laplace
from snsql.sql._mechanisms.approx_bounds import approx_bounds
from snsql.sql.privacy import Privacy
import numpy as np

class StandardScaler(CachingColumnTransformer):
    """Transforms a column of values to scale with mean centered on 0 and unit variance.
    Some privacy budget is always used to estimate the mean and variance.  If upper and lower
    are not supplied, the budget will also be used to estimate the bounds of the column.

    :param lower: The minimum value to scale to.
    :param upper: The maximum value to scale to.
    :param epsilon: The privacy budget to use.
    :param nullable: Whether the column can contain null values.  If True, the output will be a tuple of (value, null_flag).
    :param odometer: The optional privacy odometer to use to track privacy budget spent.
    """
    def __init__(self, *, lower=None, upper=None, epsilon=0.0, nullable=False, odometer=None):
        self.lower = lower
        self.upper = upper
        self.epsilon = epsilon
        self.budget_spent = []
        self.nullable = nullable
        self.odometer = odometer
        self.scaler = None
        self.mean = None
        self.var = None
        super().__init__()
    @property
    def output_type(self):
        return ColumnType.CONTINUOUS
    @property
    def needs_epsilon(self):
        return True
    @property
    def cardinality(self):
        if self.nullable:
            return [None, 2]
        else:
            return [None]
    def allocate_privacy_budget(self, epsilon, odometer):
        self.epsilon = epsilon
        self.odometer = odometer
    def _fit_finish(self):
        if self.scaler is None:
            if self.epsilon is None or self.epsilon == 0.0:
                raise ValueError("StandardScaler requires epsilon to estimate mean and variance.")
            self._fit_vals = [float(v) for v in self._fit_vals if v is not None and not (isinstance(v, float) and np.isnan(v))]
            # set bounds
            if self.upper is None or self.lower is None:
                bounds_eps = self.epsilon / 2
                self.epsilon -= bounds_eps
                self.fit_lower, self.fit_upper = approx_bounds(self._fit_vals, bounds_eps)
                if self.odometer is not None:
                    self.odometer.spend(Privacy(epsilon=bounds_eps, delta=0.0))
                self.budget_spent.append(bounds_eps)
                if self.fit_lower is None or self.fit_upper is None:
                    raise ValueError("StandardScaler could not find upper and lower bounds.")
            else:
                self.fit_lower = self.lower
                self.fit_upper = self.upper
            # fit scaler
            bounds = (float(self.fit_lower), float(self.fit_upper))
            n = len(self._fit_vals)

            enable_features("floating-point", "contrib")

            var_pre =  make_clamp(bounds) >> make_bounded_resize(n, bounds, float(self.fit_lower)) >> make_sized_bounded_variance(size=n, bounds=bounds)
            mean_pre =  make_clamp(bounds) >> make_bounded_resize(n, bounds, float(self.fit_lower)) >> make_sized_bounded_mean(size=n, bounds=bounds)

            v_e = self.epsilon * 0.8
            m_e = self.epsilon - v_e
            v_s = binary_search_param(lambda s: var_pre >> make_base_laplace(s), d_in=1, d_out=v_e)
            m_s = binary_search_param(lambda s: mean_pre >> make_base_laplace(s), d_in=1, d_out=m_e)
            dpvar = var_pre >> make_base_laplace(v_s)
            dpmean = mean_pre >> make_base_laplace(m_s)
            
            self.var = dpvar(np.array(self._fit_vals))
            self.var = np.clip(self.var, 0.001, (self.fit_upper - self.fit_lower) ** 2 / 4)
            self.mean = dpmean(np.array(self._fit_vals))
            self.mean = np.clip(self.mean, self.fit_lower, self.fit_upper)

            if self.odometer is not None:
                self.odometer.spend(Privacy(epsilon=self.epsilon, delta=0.0))
            self.budget_spent.append(self.epsilon)
        self._fit_complete = True
        if self.nullable:
            self.output_width = 2
        else:
            self.output_width = 1
    def _clear_fit(self):
        self._reset_fit()
        self.fit_lower = None
        self.fit_upper = None
    def _transform(self, val):
        if not self.fit_complete:
            raise ValueError("StandardScaler has not been fit yet.")
        if self.nullable and (val is None or isinstance(val, float) and np.isnan(val)):
            return (0.0, 1)
        else:
            val = (val - self.mean) / np.sqrt(self.var)
        if self.nullable:
            return (val, 0)
        else:
            return val
    def _inverse_transform(self, val):
        if not self.fit_complete:
            raise ValueError("StandardScaler has not been fit yet.")
        if self.nullable:
            v, n = val
            val = v
            if n == 1:
                return None
        val = val * np.sqrt(self.var) + self.mean
        return np.clip(val, self.fit_lower, self.fit_upper)



from snsynth.transform.definitions import ColumnType
from .base import ColumnTransformer
import numpy as np

class LabelTransformer(ColumnTransformer):
    """Transforms categorical values into integer-indexed labels.  Labels will be sorted if possible,
    so that the output can be used as an ordinal.  The indices will be 0-based.

    :param nullable: If null values are expected, a second output will be generated indicating null.
    """
    def __init__(self, nullable=True):
        super().__init__()
        self.nullable = nullable
    @property
    def output_type(self):
        return ColumnType.CATEGORICAL
    @property
    def cardinality(self):
        return [len(self.categories)]
    def _fit(self, val):
        if isinstance(val, float) and np.isnan(val):
            val = None
        if val not in self.labels:
            self.labels[val] = self.category
            self.categories[self.category] = val
            self.category += 1
            self.output_width = 1
    def _fit_finish(self):
        self._fit_complete = True

        # try sorting the categories so this can be used in ordinals
        vals = [v for v in self.labels.keys() if v is not None]
        val_types = set([type(v) for v in vals])
        if len(val_types) > 1:
            return
        sorted_labels = sorted(vals)
        self.labels = {}
        self.categories = []
        for i, label in enumerate(sorted_labels):
            self.labels[label] = i
            self.categories.append(label)
        if self.nullable:
            idx = len(self.categories)
            self.labels[None] = idx
            self.categories.append(None)
        return

    def _clear_fit(self):
        self._reset_fit()
        self.labels = {}
        self.categories = {}
        self.category = 0
    def _transform(self, val):
        if isinstance(val, float) and np.isnan(val):
            val = None
        return self.labels[val]
    def _inverse_transform(self, val):
        if val is None and self.nullable:
            return None
        return self.categories[val]



import torch.nn as nn


class Discriminator(nn.Module):
    def __init__(self, input_dim, wasserstein=False):
        super(Discriminator, self).__init__()
        self.model = nn.Sequential(
            nn.Linear(input_dim, 2 * input_dim // 3),
            nn.LeakyReLU(0.2),
            nn.Linear(2 * input_dim // 3, input_dim // 3),
            nn.LeakyReLU(0.2),
            nn.Linear(input_dim // 3, 1),
        )

        if not wasserstein:
            self.model.add_module("activation", nn.Sigmoid())

    def forward(self, x):
        return self.model(x)



from .aggregate_seeded import (
    AggregateSeededSynthesizer,
    AccuracyMode,
    FabricationMode,
    AggregateSeededDataset,
)


__all__ = [
    "AggregateSeededSynthesizer",
    "AccuracyMode",
    "FabricationMode",
    "AggregateSeededDataset",
]



import numpy as np
import datetime
from snsynth.transform.definitions import ColumnType
from .base import ColumnTransformer

class DateTimeTransformer(ColumnTransformer):
    """Converts a datetime column into an float column with number of days since the epoch.

    :param epoch: The epoch to use.  If None, the standard epoch of Jan 1, 1970 will be used.
    """
    def __init__(self, epoch=None):
        super().__init__()
        if epoch is None:
            self.epoch = datetime.datetime(1970, 1, 1)
        elif isinstance(epoch, str):
            self.epoch = None
            epoch = self._parse_date(epoch)
            if epoch is None:
                raise ValueError(f"Invalid epoch: {epoch}")
            self.epoch = epoch
        elif isinstance(epoch, datetime.date):
            self.epoch = datetime.datetime(epoch.year, epoch.month, epoch.day)
        elif isinstance(epoch, datetime.datetime):
            self.epoch = epoch
        else:
            raise ValueError(f"Invalid epoch: {epoch}.")
        self.format = "datetime" # can also be date or time, will be inferred from data
        self.string = False
    @property
    def output_type(self):
        return ColumnType.CONTINUOUS
    @property
    def cardinality(self):
        return [None]
    def _fit(self, val, idx=None):
        pass
    def _clear_fit(self):
        # this transform doesn't need fit
        self._fit_complete = True
        self.output_width = 1
    def _transform(self, val):
        if val is None or (isinstance(val, float) and np.isnan(val)):
            return np.nan
        else:
            parsed = self._parse_date(val)
            if parsed is None:
                raise ValueError(f"Invalid date: {val}")
            distance = parsed - self.epoch
            return float(distance.total_seconds() / (60 * 60 * 24))
    def _inverse_transform(self, val):
        if val is None or (isinstance(val, float) and np.isnan(val)):
            return None
        else:
            new_date = self.epoch + datetime.timedelta(days=val)
            if self.string:
                if self.format == "datetime":
                    return new_date.isoformat()
                elif self.format == "date":
                    return new_date.date().isoformat()
                elif self.format == "time":
                    return new_date.time().isoformat()
            else:
                if self.format == "datetime":
                    return new_date
                elif self.format == "date":
                    return new_date.date()
                elif self.format == "time":
                    return new_date.time()
                else:
                    raise ValueError(f"Invalid format: {self.format}")
    def _parse_date(self, val):
        if val is None:
            return None
        if isinstance(val, datetime.datetime):
            self.format = "datetime"
            self.string = False
            return val
        elif isinstance(val, datetime.date):
            self.format = "date"
            self.string = False
            return datetime.datetime(val.year, val.month, val.day)
        elif isinstance(val, datetime.time):
            self.format = "time"
            self.string = False
            return datetime.datetime.combine(self.epoch, val)
        elif isinstance(val, str):
            self.string = True
            try:
                parsed = datetime.datetime.strptime(val, "%Y-%m-%d")
                self.format = "date"
                return parsed
            except:
                pass
            try:
                parsed = datetime.datetime.strptime(val, "%H:%M:%S").time()
                self.format = "time"
                if self.epoch is None:
                    self.epoch = datetime.datetime(1970, 1, 1)
                return datetime.datetime.combine(self.epoch, parsed)
            except:
                pass
            try:
                parsed = datetime.datetime.strptime(val, "%H:%M:%S.%f").time()
                self.format = "time"
                if self.epoch is None:
                    self.epoch = datetime.datetime(1970, 1, 1)
                return datetime.datetime.combine(self.epoch, parsed)
            except:
                pass
            try:
                parsed = datetime.datetime.strptime(val, "%I:%M:%S %p").time()
                self.format = "time"
                if self.epoch is None:
                    self.epoch = datetime.datetime(1970, 1, 1)
                return datetime.datetime.combine(self.epoch, parsed)
            except:
                pass
            try:
                parsed = datetime.datetime.strptime(val, "%I:%M:%S.%f %p").time()
                self.format = "time"
                if self.epoch is None:
                    self.epoch = datetime.datetime(1970, 1, 1)
                return datetime.datetime.combine(self.epoch, parsed)
            except:
                pass
            try:
                parsed = datetime.datetime.fromisoformat(val)
                self.format = "datetime"
                return parsed
            except:
                pass
            try:
                parsed = datetime.datetime.strptime(val, "%Y-%m-%d %H:%M:%S")
                self.format = "datetime"
                return parsed
            except:
                pass
            try:
                parsed = datetime.datetime.strptime(val, "%Y-%m-%d %H:%M:%S.%f")
                self.format = "datetime"
                return parsed
            except:
                pass
            try:
                parsed = datetime.datetime.strptime(val, "%Y-%m-%d %I:%M:%S %p")
                self.format = "datetime"
                return parsed
            except:
                pass
            try:
                parsed = datetime.datetime.strptime(val, "%Y-%m-%d %I:%M:%S.%f %p")
                self.format = "datetime"
                return parsed
            except:
                pass
            try:
                parsed = datetime.datetime.strptime(val, "%Y-%m-%dT%H:%M:%S")
                self.format = "datetime"
                return parsed
            except:
                pass
            try:
                parsed = datetime.datetime.strptime(val, "%Y-%m-%dT%H:%M:%S.%f")
                self.format = "datetime"
                return parsed
            except:
                pass
            try:
                parsed = datetime.datetime.strptime(val, "%Y-%m-%dT%H:%M:%S%z")
                self.format = "datetime"
                return parsed
            except:
                pass
            try:
                parsed = datetime.datetime.strptime(val, "%Y-%m-%dT%H:%M:%S.%f%z")
                self.format = "datetime"
                return parsed
            except:
                pass
            return None
        else:
            return None



from .base import CachingColumnTransformer
from snsql.sql._mechanisms.approx_bounds import approx_bounds
from snsql.sql.privacy import Privacy
from snsynth.transform.definitions import ColumnType
import numpy as np

class BinTransformer(CachingColumnTransformer):
    """Transforms continuous values into a discrete set of bins.

    :param bins: The number of bins to create.
    :param lower: The minimum value to scale to.
    :param upper: The maximum value to scale to.
    :param epsilon: The privacy budget to use to infer bounds, if none provided.
    :param nullable: If null values are expected, a second output will be generated indicating null.
    :param odometer: The optional odometer to use to track privacy budget.
    """
    def __init__(self, *, bins=10, lower=None, upper=None, epsilon=0.0, nullable=False, odometer=None):
        self.lower = lower
        self.upper = upper
        self.epsilon = epsilon
        self.bins = bins
        self.budget_spent = []
        self.nullable = nullable
        self.odometer = odometer
        super().__init__()
    @property
    def output_type(self):
        return ColumnType.CATEGORICAL
    @property
    def needs_epsilon(self):
        return self.upper is None or self.lower is None
    @property
    def cardinality(self):
        if self.nullable:
            return [self.bins, 2]
        else:
            return [self.bins]
    def allocate_privacy_budget(self, epsilon, odometer):
        self.epsilon = epsilon
        self.odometer = odometer
    def _fit_finish(self):
        if self.epsilon is not None and self.epsilon > 0.0 and (self.lower is None or self.upper is None):
            self._fit_vals = [v for v in self._fit_vals if v is not None and not (isinstance(v, float) and np.isnan(v))]
            self.fit_lower, self.fit_upper = approx_bounds(self._fit_vals, self.epsilon)
            if self.odometer is not None:
                self.odometer.spend(Privacy(epsilon=self.epsilon, delta=0.0))
            self.budget_spent.append(self.epsilon)
            if self.fit_lower is None or self.fit_upper is None:
                raise ValueError("BinTransformer could not find bounds.")
        elif self.lower is None or self.upper is None:
            raise ValueError("BinTransformer requires either epsilon or min and max.")
        else:
            self.fit_lower = self.lower
            self.fit_upper = self.upper
        self._fit_complete = True
        if self.nullable:
            self.output_width = 2
        else:
            self.output_width = 1
    def _clear_fit(self):
        self._reset_fit()
        self.fit_lower = None
        self.fit_upper = None
        # if bounds provided, we can immediately use without fitting
        if self.lower and self.upper:
            self._fit_complete = True
            if self.nullable:
                self.output_width = 2
            else:
                self.output_width = 1
            self.fit_lower = self.lower
            self.fit_upper = self.upper
    def _bin_edges(self, bin):
        return (
            self.fit_lower + (bin / self.bins) * (self.fit_upper - self.fit_lower),
            self.fit_lower + ((bin + 1) / self.bins) * (self.fit_upper - self.fit_lower)
        )
    def _bin(self, val):
        if not self.fit_complete:
            raise ValueError("BinTransformer has not been fit yet.")
        if self.nullable and (val is None or (isinstance(val, float) and np.isnan(val))):
            return 1
        return int(self.bins * (val - self.fit_lower) / (self.fit_upper - self.fit_lower))
    def _transform(self, val):
        if not self.fit_complete:
            raise ValueError("BinTransformer has not been fit yet.")
        if  val is None or (isinstance(val, float) and np.isnan(val)):
            if self.nullable:
                return (0, 1)
            else:
                raise ValueError("Cannot transform None or NaN.  Consider setting nullable=True.")
        val = self.fit_lower if val < self.fit_lower else val
        val = self.fit_upper if val > self.fit_upper else val
        if self.nullable:
            return (self._bin(val), 0)
        else:
            return self._bin(val)
    def _inverse_transform(self, val):
        if not self.fit_complete:
            raise ValueError("BinTransformer has not been fit yet.")
        if self.nullable:
            v, n = val
            if n == 1 or v is None or (isinstance(v, float) and np.isnan(v)):
                return None
            val = v
        lower, upper = self._bin_edges(val)
        return (lower + upper) / 2



import torch


class BaseSynthesizer:
    """Base class for all default synthesizers of ``CTGAN``.

    This should contain the save/load methods.
    """

    def save(self, path):
        device_backup = self._device
        self.set_device(torch.device("cpu"))
        torch.save(self, path)
        self.set_device(device_backup)

    @classmethod
    def load(cls, path):
        device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        model = torch.load(path)
        model.set_device(device)
        return model


from .dpgan import DPGAN
from .pategan import PATEGAN


# try:
from .dpctgan import DPCTGAN  # noqa
from .patectgan import PATECTGAN  # noqa

__all__ = ["DPCTGAN", "PATECTGAN", "DPGAN", "PATEGAN"]
# except ImportError as e:
#     import logging

#     logger = logging.getLogger(__name__)
#     logger.warning(
#         "Requires: pip install 'ctgan==0.2.2.dev1' for ctgan based synthesizers. Failed with exception {}".format(e))

#     __all__ = ["DPGAN", "PATEGAN"]




from snsynth.transform.definitions import ColumnType


class ColumnTransformer:
    """Base class for column transformers.  Subclasses must implement the
    _fit, _transform, and _inverse_transform methods."""
    def __init__(self):
        self._fit_complete = False
        self.output_width = 0
        self._clear_fit()
    @property
    def output_type(self):
        """Must be implemented by subclasses to return the type of the output
        of the transformer.  From definitions.ColumnType enum."""
        raise NotImplementedError
    @property
    def is_categorical(self):
        return self.output_type == ColumnType.CATEGORICAL
    @property
    def is_continuous(self):
        return self.output_type == ColumnType.CONTINUOUS
    @property
    def fit_complete(self):
        return self._fit_complete
    @property
    def needs_epsilon(self):
        """Overriden by subclasses to indicate whether the transformer
        needs an epsilon value to be supplied to the fit method."""
        return False
    @property
    def cardinality(self):
        """Must be implemented by subclasses to return the cardinality of the
        output of the transformer.  Only applicable for categorical transformers."""
        return []
    def allocate_privacy_budget(self, epsilon, odometer):
        """Allocate privacy budget to the transformer.  This method is called
        by the DataTransformer to allocate privacy budget to the transformer.
        The default implementation does nothing, but subclasses can override
        this method to allocate privacy budget to the transformer.
        """
        pass
    def fit(self, data, idx=None):
        """Fit a column of data.  If data includes multiple columns,
        provide an index to select which column to fit.  If no index is supplied,
        data must be a single column as a list of values.

        .. code-block:: python

            t = MyColumnTransformer()
            assert(t.fit_complete == False)
            t.fit(iris, 0)  # fit the first column of iris
            assert(t.fit_complete == True)

        This method will always consume all rows in the data, even if the fit method
        has already been called.  This is to ensure consistency when passing iterators.
        If you want to ensure that fit only gets called when the transformer has not already
        been fit, you can check the fit_complete property.

        .. code-block:: python

            t = MyColumnTransformer()
            if not t.fit_complete:
                t.fit(iris, 0)
            assert(t.fit_complete == True)

        Calling this method repeatedly should perform a fresh fit each time using the new data.  
        """
        self._clear_fit()
        if idx is None:
            for val in data:
                self._fit(val)
        else:
            for row in data:
                self._fit(row[idx])
        self._fit_finish()
    def transform(self, data, idx=None):
        """Transform a column of data.  If data includes multiple columns,
        provide an index to select which column to transform.  If no index is supplied,
        data must be a single column as a list of values.
        
        .. code-block:: python

            t = MyColumnTransformer()
            t.fit(iris, 0)  # fit the first column of iris
            iris_encoded = t.transform(iris, 0)  # transform the first column of iris
        """
        if idx is None:
            return [self._transform(val) for val in data]
        else:
            return [self._transform(row[idx]) for row in data]
    def fit_transform(self, data, idx=None):
        """Fit and transform a column of data.  If data includes multiple columns,
        provide an index to select which column to fit.  If no index is supplied,
        data must be a single column as a list of values.

        .. code-block:: python

            t = MyColumnTransformer()
            iris_encoded = t.fit_transform(iris, 0)  # fit and transform the first column of iris

        This method will always perform a fresh fit each time it is called.  A fit and a transform
        will each consume all rows in the data, so this method should be called only on a data structure
        that allows multiple passes over the data.
        """
        if not self.fit_complete:
            self.fit(data, idx)
        return self.transform(data, idx)
    def inverse_transform(self, data, idx=None):
        """Inverse-transform a column of data.  If data includes multiple columns,
        provide an index to select which column to inverse-transform.  If no index is supplied,
        data must be a single column as a list of values.
        
        .. code-block:: python

            t = MyColumnTransformer()
            t.fit(iris, 0)  # fit the first column of iris
            iris_encoded = t.transform(iris, 0)  # transform the first column of iris
            iris_decoded = t.inverse_transform(iris_encoded, 0)  # inverse-transform the first column of iris
            assert(all([a == b for a, b in zip(iris[0], iris_decoded)]))
            
        """
        if idx is None:
            return [self._inverse_transform(val) for val in data]
        else:
            return [self._inverse_transform(row[idx]) for row in data]
    def _fit(self, val):
        """Must be implemented by subclasses to fit a single value.
        """
        raise NotImplementedError
    def _fit_finish(self):
        """Should be implemented by subclasses if there is
        any work that needs to be done to finish the fit process."""
        self._fit_complete = True
    def _clear_fit(self):
        """Must be implemented by subclasses to clear any state that is
        dependent on the data that was fit.
        """
        raise NotImplementedError
    def _reset_fit(self):
        """Helper can be called by _clear_fit"""
        self._fit_complete = False
        self.output_width = 0
    def _transform(self, val):
        """Must be implemented by subclasses to transform a single value."""
        raise NotImplementedError
    def _inverse_transform(self, val):
        """Must be implemented by subclasses to inverse-transform a single value."""
        raise NotImplementedError

class CachingColumnTransformer(ColumnTransformer):
    """Base class for column transformers that cache their fit state.
    Subclasses must implement the _fit_finish, _transform and _inverse_transform methods.
    """
    def __init__(self):
        super().__init__()
        self._fit_vals = []
    def _fit(self, val):
        """Caches each fit value to be processed by _fit_finish.
        If this method is overridden, it should call super()._fit(val)"""
        self._fit_vals.append(val)
    def _reset_fit(self):
        """Helper can be called by _clear_fit"""
        super()._reset_fit()
        self._fit_vals = []



from snsynth.transform.definitions import ColumnType
from .base import ColumnTransformer
import numpy as np

class OneHotEncoder(ColumnTransformer):
    """Transforms integer-labeled data into one-hot encoding.  Inputs are assumed to be 0-based.
    To convert from unstructured categorical data, chain with LabelTransformer first.
    """
    cache_fit = False
    def __init__(self):
        super().__init__()
    @property
    def output_type(self):
        return ColumnType.CATEGORICAL
    @property
    def cardinality(self):
        return [2] * (self.max + 1)
    def _fit(self, val):
        if val > self.max:
            self.max = val
    def _fit_finish(self):
        self.output_width = self.max + 1
        super()._fit_finish()
    def _clear_fit(self):
        self._fit_complete = False
        self.max = -1
    def _transform(self, val):
        if self.max < 0 or not self._fit_complete:
            raise ValueError("OneHotEncoder has not been fit yet.")
        elif val < 0 or val > self.max:
            raise ValueError(
                f"Provided integer-label {val} is invalid."
                " Please ensure that all inputs are 0-based and provided during data fit."
            )
        elif self.max == 0:
            return 1

        bits = [0] * (self.max + 1)
        bits[val] = 1
        return tuple(bits)
    def _inverse_transform(self, val):
        # will always choose first if multiple are set
        return np.argmax(val)



import pandas as pd
import numpy as np

from snsynth.transform.table import TableTransformer
from snsql.sql.parse import QueryParser


class SDGYMBaseSynthesizer:
    def fit(
        self, 
        data, 
        *ignore, 
        transformer=None, 
        categorical_columns=[], 
        ordinal_columns=[], 
        continuous_columns=[],
        preprocessor_eps=0.0,
        nullable=False
        ):
        """
        Fit the synthesizer model on the data.

        :param data: The private data used to fit the synthesizer.
        :type data: pd.DataFrame, np.ndarray, or list of tuples
        :param transformer: The transformer to use to preprocess the data.
            If no transformer is provided, the synthesizer will attempt to choose a transformer suitable for that synthesizer.
            To prevent the synthesizer from choosing a transformer, pass in snsynth.transform.NoTransformer().
            The inferred preprocessor can be constrained for certain columns by providing a dictionary.
            Read the ``TableTransformer.create()`` method documentation for details about the constraints.
        :type transformer: snsynth.transform.TableTransformer or dictionary, optional
        :param categorical_columns: List of column names or indixes to be treated as categorical columns, used as hints when no transformer is provided.
        :type categorical_columns: list[], optional
        :param ordinal_columns: List of column names or indices to be treated as ordinal columns, used as hints when no transformer is provided.
        :type ordinal_columns: list[], optional
        :param continuous_columns: List of column names or indices to be treated as continuous columns, used as hints when no transformer is provided.
        :type continuous_columns: list[], optional
        :param preprocessor_eps: The epsilon value to use when preprocessing the data.  This epsilon budget is subtracted from the
            budget supplied when creating the synthesizer, but is only used if the preprocessing requires
            privacy budget, for example if bounds need to be inferred for continuous columns.  This value defaults to
            0.0, and the synthesizer will raise an error if the budget is not sufficient to preprocess the data.
        :type preprocessor_eps: float, optional
        :param nullable: Whether to allow null values in the data. This is only used if no transformer is provided,
            and is used as a hint when inferring transformers. Defaults to False.
        :type nullable: bool, optional
        """
        raise NotImplementedError

    def sample(self, n_rows):
        """
        Sample rows from the synthesizer.

        :param n_rows: The number of rows to create
        :type n_rows: int
        :return: Data set containing the generated data samples.
        :rtype: pd.DataFrame, np.ndarray, or list of tuples
        """
        raise NotImplementedError

    def fit_sample(
        self, 
        data, *ignore, 
        transformer=None, 
        categorical_columns=[], 
        ordinal_columns=[], 
        continuous_columns=[],
        preprocessor_eps=0.0,
        nullable=False,
        **kwargs
        ):
        """
        Fit the synthesizer model and then generate a synthetic dataset of the same
        size of the input data.

        :param data: The private data used to fit the synthesizer.
        :type data: pd.DataFrame, np.ndarray, or list of tuples
        :param transformer: The transformer to use to preprocess the data.
            If no transformer is provided, the synthesizer will attempt to choose a transformer suitable for that synthesizer.
            To prevent the synthesizer from choosing a transformer, pass in snsynth.transform.NoTransformer().
            The inferred preprocessor can be constrained for certain columns by providing a dictionary.
            Read the ``TableTransformer.create()`` method documentation for details about the constraints.
        :type transformer: snsynth.transform.TableTransformer or dict, optional
        :param categorical_columns: List of column names or indixes to be treated as categorical columns, used as hints when no transformer is provided.
        :type categorical_columns: list[], optional
        :param ordinal_columns: List of column names or indices to be treated as ordinal columns, used as hints when no transformer is provided.
        :type ordinal_columns: list[], optional
        :param continuous_columns: List of column names or indices to be treated as continuous columns, used as hints when no transformer is provided.
        :type continuous_columns: list[], optional
        :param preprocessor_eps: The epsilon value to use when preprocessing the data.  This epsilon budget is subtracted from the
            budget supplied when creating the synthesizer, but is only used if the preprocessing requires
            privacy budget, for example if bounds need to be inferred for continuous columns.  This value defaults to
            0.0, and the synthesizer will raise an error if the budget is not sufficient to preprocess the data.
        :type preprocessor_eps: float, optional
        :param nullable: Whether to allow null values in the data. This is only used if no transformer is provided,
            and is used as a hint when inferring transformers. Defaults to False.
        :type nullable: bool, optional
        """
        self.fit(
            data, 
            transformer=transformer, 
            categorical_columns=categorical_columns, 
            ordinal_columns=ordinal_columns, 
            continuous_columns=continuous_columns,
            preprocessor_eps=preprocessor_eps,
            nullable=nullable,
            **kwargs
        )
        if isinstance(data, pd.DataFrame):
            return self.sample(len(data))
        elif isinstance(data, np.ndarray):
            return self.sample(data.shape[0])
        elif isinstance(data, list):
            return self.sample(len(data))
        else:
            raise TypeError('Data must be a pandas DataFrame, numpy array, or list of tuples')

synth_map = {
    'mwem': {
        'class': 'snsynth.mwem.MWEMSynthesizer'
    },
    'dpctgan' : {
        'class': 'snsynth.pytorch.nn.dpctgan.DPCTGAN'
    },
    'patectgan' : {
        'class': 'snsynth.pytorch.nn.patectgan.PATECTGAN'
    },
    'mst': {
        'class': 'snsynth.mst.mst.MSTSynthesizer'
    },
    'pacsynth': {
        'class': 'snsynth.aggregate_seeded.AggregateSeededSynthesizer'
    },
    'dpgan': {
        'class': 'snsynth.pytorch.nn.dpgan.DPGAN'
    },
    'pategan': {
        'class': 'snsynth.pytorch.nn.pategan.PATEGAN'
    }
}

class Synthesizer(SDGYMBaseSynthesizer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    @classmethod
    def list_synthesizers(cls):
        """
        List the available synthesizers.
        
        :return: List of available synthesizer names.
        :rtype: list[str]
        """
        return list(synth_map.keys())
    def _get_train_data(self, data, *ignore, style, transformer, categorical_columns, ordinal_columns, continuous_columns, nullable, preprocessor_eps):
        if transformer is None or isinstance(transformer, dict):
            self._transformer = TableTransformer.create(data, style=style,
                categorical_columns=categorical_columns,
                continuous_columns=continuous_columns,
                ordinal_columns=ordinal_columns,
                nullable=nullable,
                constraints=transformer)
        elif isinstance(transformer, TableTransformer):
            self._transformer = transformer
        else:
            raise ValueError("transformer must be a TableTransformer object, a dictionary or None.")
        if not self._transformer.fit_complete:
            if self._transformer.needs_epsilon and (preprocessor_eps is None or preprocessor_eps == 0.0):
                raise ValueError("Transformer needs some epsilon to infer bounds.  If you know the bounds, pass them in to save budget.  Otherwise, set preprocessor_eps to a value > 0.0 and less than the training epsilon.  Preprocessing budget will be subtracted from training budget.")
            self._transformer.fit(
                data,
                epsilon=preprocessor_eps
            )
            eps_spent, _ = self._transformer.odometer.spent
            if eps_spent > 0.0:
                self.epsilon -= eps_spent
                print(f"Spent {eps_spent} epsilon on preprocessor, leaving {self.epsilon} for training")
                if self.epsilon < 10E-3:
                    raise ValueError("Epsilon remaining is too small!")
        train_data = self._transformer.transform(data)
        return train_data

    # factory method
    @classmethod
    def create(cls, synth=None, epsilon=None, *args, **kwargs):
        """
        Create a differentially private synthesizer.

        :param synth: The name of the synthesizer to create.  If called from an instance of a Synthesizer subclass, creates
            an instance of the specified synthesizer.  Allowed synthesizers are available from
            the list_synthesizers() method.
        :type synth: str or Synthesizer class, required
        :param epsilon: The privacy budget to be allocated to the synthesizer.  This budget will be
            used when the synthesizer is fit to the data.
        :type epsilon: float, required
        :param args: Positional arguments to pass to the synthesizer constructor.
        :type args: list, optional
        :param kwargs: Keyword arguments to pass to the synthesizer constructor.  At a minimum,
            the epsilon value must be provided.  Any other hyperparameters can be provided
            here.  See the documentation for each specific synthesizer for details about available
            hyperparameter.
        :type kwargs: dict, optional

        """
        if synth is None or (isinstance(synth, type) and issubclass(synth, Synthesizer)):
            clsname = cls.__module__ + '.' + cls.__name__ if synth is None else synth.__module__ + '.' + synth.__name__
            if clsname == 'snsynth.base.Synthesizer':
                raise ValueError("Must specify a synthesizer to use.")
            matching_keys = [k for k, v in synth_map.items() if v['class'] == clsname]
            if len(matching_keys) == 0:
                raise ValueError(f"Synthesizer {clsname} not found in map.")
            elif len(matching_keys) > 1:
                raise ValueError(f"Synthesizer {clsname} found multiple times in map.")
            else:
                synth = matching_keys[0]
        if isinstance(synth, str):
            synth = synth.lower()
            if synth not in synth_map:
                raise ValueError('Synthesizer {} not found'.format(synth))
            synth_class = synth_map[synth]['class']
            synth_module, synth_class = synth_class.rsplit('.', 1)
            synth_module = __import__(synth_module, fromlist=[synth_class])
            synth_class = getattr(synth_module, synth_class)
            return synth_class(epsilon=epsilon, *args, **kwargs)
        else:
            raise ValueError('Synthesizer must be a string or a class')

    @staticmethod
    def _evaluate_condition(samples, query_condition, valid_samples, column_names):
        """
        Evaluates every sample with the given condition.
        If a sample is valid, it will be appended to the list of valid samples.

        :param samples: Data set containing generated data samples.
        :type samples: iterable, required
        :param query_condition: Condition to evaluate.
        :type query_condition: snsql._ast.tokens.SqlExpr, required
        :param valid_samples: List of valid data samples.
        :type valid_samples: list[pd.Series, np.ndarray, or tuples], required
        :param column_names: List of column names, used if ``samples`` is not a pd.DataFrame.
        :type column_names: list[str], required
        """
        if isinstance(samples, pd.DataFrame):
            for _, sample in samples.iterrows():
                if query_condition.evaluate(sample):
                    valid_samples.append(sample)
        else:
            for sample in samples:
                try:  # check if the sample is iterable
                    iter(sample)
                except TypeError:  # else wrap the single value in a list
                    sample = [sample]
                bindings = dict(zip(column_names, sample))
                if query_condition.evaluate(bindings):
                    valid_samples.append(sample)

    def sample_conditional(self, n_rows, condition, max_tries=100, column_names=None):
        """
        Generates a synthetic dataset that satisfies the given condition.
        Performs a rejection sampling process where up to ``max_tries`` batches are sampled.
        If not enough valid data samples are found the partial dataset is returned.

        :param n_rows: Number of rows to sample.
        :type n_rows: int, required
        :param condition: Condition to evaluate. Needs to be a valid SQL WHERE clause e.g. "age < 50 AND income > 1000".
        :type condition: str, required
        :param max_tries: Number of times to retry sampling until enough are generated. Defaults to 100.
        :type max_tries: int, optional
        :param column_names: List of column names, required if ``samples`` is not a pd.DataFrame.
        :type column_names: list[str], optional
        :return: Data set containing the generated data samples.
        :rtype: pd.DataFrame, np.ndarray, or list of tuples
        """
        if n_rows < 1:
            raise ValueError(f"Please provide a value >= 1 for `n_rows`")

        if max_tries < 1:
            raise ValueError(f"Please provide a value >= 1 for `max_tries`")

        try:
            query = QueryParser().query(f"SELECT * FROM DUMMY WHERE {condition}")
            query_condition = query.where.condition
        except ValueError as e:
            raise ValueError(
                f"Could not parse `condition` {condition}. Please provide a valid WHERE clause"
            ) from e

        samples = self.sample(n_rows)
        if not isinstance(samples, pd.DataFrame) and column_names is None:
            raise ValueError(
                f"Please provide `column_names` for samples of type {type(samples)}"
            )

        valid_samples = []
        valid_total = 0
        n_sample = n_rows
        tries = 1

        while valid_total < n_rows:
            valid_before = valid_total
            try:
                self._evaluate_condition(
                    samples, query_condition, valid_samples, column_names
                )
            except ValueError as e:
                raise ValueError(
                    f"Could not evaluate `condition` {condition}. Please make sure to use valid column names"
                ) from e

            valid_total = len(valid_samples)

            valid_current = valid_total - valid_before
            valid_rate = max(valid_current, 1) / max(
                n_sample, 1
            )  # can decrease unnecessary sampling

            remaining = n_rows - valid_total
            n_sample = min(10 * n_rows, int(remaining / valid_rate))

            tries += 1
            if tries >= max_tries or n_sample < 1:
                break
            samples = self.sample(n_sample)

        max_length = min(valid_total, n_rows)
        valid_samples = valid_samples[:max_length]
        if isinstance(samples, pd.DataFrame):
            return pd.DataFrame(
                data=valid_samples, index=pd.RangeIndex(stop=max_length)
            ).astype(samples.dtypes)
        elif isinstance(samples, np.ndarray):
            return np.array(valid_samples)
        else:
            return valid_samples



from .base import ColumnTransformer

class ClampTransformer(ColumnTransformer):
    """Clamps values to be within a specified range.

    :param lower: The minimum value to scale to.  If None, no lower bound is applied.
    :param upper: The maximum value to scale to.  If None, no upper bound is applied.
    """
    def __init__(self, upper=None, lower=None):
        super().__init__()
        if upper is None and lower is None:
            raise ValueError('Must specify upper and/or lower')
        self.upper = upper
        self.lower = lower
    @property
    def cardinality(self):
        return [1]
    def _fit(self, val):
        pass
    def _clear_fit(self):
        self._fit_complete = True
        self.output_width = 1
    def _transform(self, val):
        if self.upper is not None and val > self.upper:
            return self.upper
        if self.lower is not None and val < self.lower:
            return self.lower
        return val
    def _inverse_transform(self, val):
        if self.upper is not None and val > self.upper:
            return self.upper
        if self.lower is not None and val < self.lower:
            return self.lower
        return val



"""
   Copyright 2020 (https://github.com/IBM/discrete-gaussian-differential-privacy)

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

# Code for computing approximate differential privacy guarantees
# for discrete Gaussian and, more generally, concentrated DP
# See https://arxiv.org/abs/2004.00010
# - Thomas Steinke dgauss@thomas-steinke.net 2020

import math

# *********************************************************************
# Now we move on to concentrated DP

# compute delta such that
# rho-CDP implies (eps,delta)-DP
# Note that adding cts or discrete N(0,sigma2) to sens-1 gives rho=1/(2*sigma2)


# start with standard P[privloss>eps] bound via markov
def cdp_delta_standard(rho, eps):
    assert rho >= 0
    assert eps >= 0
    if rho == 0:
        return 0  # degenerate case
    # https://arxiv.org/pdf/1605.02065.pdf#page=15
    return math.exp(-((eps - rho)**2) / (4 * rho))


# Our new bound:
# https://arxiv.org/pdf/2004.00010v3.pdf#page=13
def cdp_delta(rho, eps):
    assert rho >= 0
    assert eps >= 0
    if rho == 0:
        return 0  # degenerate case

    # search for best alpha
    # Note that any alpha in (1,infty) yields a valid upper bound on delta
    # Thus if this search is slightly "incorrect" it will only result in larger delta (still valid)
    # This code has two "hacks".
    # First the binary search is run for a pre-specificed length.
    # 1000 iterations should be sufficient to converge to a good solution.
    # Second we set a minimum value of alpha to avoid numerical stability issues.
    # Note that the optimal alpha is at least (1+eps/rho)/2. Thus we only hit this constraint
    # when eps<=rho or close to it. This is not an interesting parameter regime, as you will
    # inherently get large delta in this regime.

    amin = 1.01  # don't let alpha be too small, due to numerical stability
    amax = (eps + 1) / (2 * rho) + 2
    for i in range(1000):  # should be enough iterations
        alpha = (amin + amax) / 2
        derivative = (2 * alpha - 1) * rho - eps + math.log1p(-1.0 / alpha)
        if derivative < 0:
            amin = alpha
        else:
            amax = alpha

    # now calculate delta
    delta = math.exp((alpha - 1) * (alpha * rho - eps) + alpha * math.log1p(-1 / alpha)) / (alpha - 1.0)
    return min(delta, 1.0)  # delta<=1 always


# Above we compute delta given rho and eps, now we compute eps instead
# That is we wish to compute the smallest eps such that rho-CDP implies (eps,delta)-DP
def cdp_eps(rho, delta):
    assert rho >= 0
    assert delta > 0
    if delta >= 1 or rho == 0:
        return 0.0  # if delta>=1 or rho=0 then anything goes
    epsmin = 0.0  # maintain cdp_delta(rho,eps)>=delta
    epsmax = rho + 2 * math.sqrt(rho * math.log(1 / delta))  # maintain cdp_delta(rho,eps)<=delta

    # to compute epsmax we use the standard bound
    for i in range(1000):
        eps = (epsmin + epsmax) / 2
        if cdp_delta(rho, eps) <= delta:
            epsmax = eps
        else:
            epsmin = eps
    return epsmax


# Now we compute rho
# Given (eps,delta) find the smallest rho such that rho-CDP implies (eps,delta)-DP
def cdp_rho(eps, delta):
    assert eps >= 0
    assert delta > 0
    if delta >= 1:
        return 0.0  # if delta>=1 anything goes
    rhomin = 0.0  # maintain cdp_delta(rho,eps)<=delta
    rhomax = eps + 1  # maintain cdp_delta(rhomax,eps)>delta
    for i in range(1000):
        rho = (rhomin + rhomax) / 2
        if cdp_delta(rho, eps) <= delta:
            rhomin = rho
        else:
            rhomax = rho
    return rhomin



import torch.nn as nn


class Generator(nn.Module):
    def __init__(self, latent_dim, output_dim, binary=True):
        super(Generator, self).__init__()

        def block(in_, out, activation):
            return nn.Sequential(nn.Linear(in_, out, bias=False), nn.LayerNorm(out), activation(),)

        self.layer_0 = block(
            latent_dim, latent_dim, nn.Tanh if binary else lambda: nn.LeakyReLU(0.2)
        )
        self.layer_1 = block(
            latent_dim, latent_dim, nn.Tanh if binary else lambda: nn.LeakyReLU(0.2)
        )
        self.layer_2 = block(
            latent_dim, output_dim, nn.Tanh if binary else lambda: nn.LeakyReLU(0.2)
        )

    def forward(self, noise):
        noise = self.layer_0(noise) + noise
        noise = self.layer_1(noise) + noise
        noise = self.layer_2(noise)
        return noise



import torch
import torch.nn as nn
import math
import numpy as np


def weights_init(m):
    if type(m) == nn.Linear:
        nn.init.xavier_uniform_(m.weight)


def pate(data, teachers, lap_scale, device="cpu"):
    """PATE implementation for GANs.
    """
    num_teachers = len(teachers)
    labels = torch.Tensor(num_teachers, data.shape[0]).type(torch.int64).to(device)
    for i in range(num_teachers):
        output = teachers[i](data)
        pred = (output > 0.5).type(torch.Tensor).squeeze().to(device)
        # print(pred.shape)
        # print(labels[i].shape)
        labels[i] = pred

    votes = torch.sum(labels, dim=0).unsqueeze(1).type(torch.DoubleTensor).to(device)
    noise = torch.from_numpy(np.random.laplace(loc=0, scale=1 / lap_scale, size=votes.size())).to(
        device
    )
    noisy_votes = votes + noise
    noisy_labels = (noisy_votes > num_teachers / 2).type(torch.DoubleTensor).to(device)

    return noisy_labels, votes


def moments_acc(num_teachers, votes, lap_scale, l_list, device="cpu"):
    q = (2 + lap_scale * torch.abs(2 * votes - num_teachers)) / (
        4 * torch.exp(lap_scale * torch.abs(2 * votes - num_teachers))
    ).to(device)

    alpha = []
    for l_val in l_list:
        a = 2 * lap_scale ** 2 * l_val * (l_val + 1)
        t_one = (1 - q) * torch.pow((1 - q) / (1 - math.exp(2 * lap_scale) * q), l_val)
        t_two = q * torch.exp(2 * lap_scale * l_val)
        t = t_one + t_two
        alpha.append(torch.clamp(t, max=a).sum())

    return torch.DoubleTensor(alpha).to(device)



from multiprocessing.sharedctypes import Value
import warnings

import numpy as np
import pandas as pd
import torch
from packaging import version
from torch import optim
from torch.nn import BatchNorm1d, Dropout, LeakyReLU, Linear, Module, ReLU, Sequential, functional

from .data_sampler import DataSampler
from .base import BaseSynthesizer
from snsynth.transform.table import TableTransformer


class Discriminator(Module):

    def __init__(self, input_dim, discriminator_dim, pac=10):
        super(Discriminator, self).__init__()
        dim = input_dim * pac
        self.pac = pac
        self.pacdim = dim
        seq = []
        for item in list(discriminator_dim):
            seq += [Linear(dim, item), LeakyReLU(0.2), Dropout(0.5)]
            dim = item

        seq += [Linear(dim, 1)]
        self.seq = Sequential(*seq)

    def calc_gradient_penalty(self, real_data, fake_data, device='cpu', pac=10, lambda_=10):
        alpha = torch.rand(real_data.size(0) // pac, 1, 1, device=device)
        alpha = alpha.repeat(1, pac, real_data.size(1))
        alpha = alpha.view(-1, real_data.size(1))

        interpolates = alpha * real_data + ((1 - alpha) * fake_data)

        disc_interpolates = self(interpolates)

        gradients = torch.autograd.grad(
            outputs=disc_interpolates, inputs=interpolates,
            grad_outputs=torch.ones(disc_interpolates.size(), device=device),
            create_graph=True, retain_graph=True, only_inputs=True
        )[0]

        gradient_penalty = ((
            gradients.view(-1, pac * real_data.size(1)).norm(2, dim=1) - 1
        ) ** 2).mean() * lambda_

        return gradient_penalty

    def forward(self, input):
        assert input.size()[0] % self.pac == 0
        return self.seq(input.view(-1, self.pacdim))


class Residual(Module):

    def __init__(self, i, o):
        super(Residual, self).__init__()
        self.fc = Linear(i, o)
        self.bn = BatchNorm1d(o)
        self.relu = ReLU()

    def forward(self, input):
        out = self.fc(input)
        out = self.bn(out)
        out = self.relu(out)
        return torch.cat([out, input], dim=1)


class Generator(Module):

    def __init__(self, embedding_dim, generator_dim, data_dim):
        super(Generator, self).__init__()
        dim = embedding_dim
        seq = []
        for item in list(generator_dim):
            seq += [Residual(dim, item)]
            dim += item
        seq.append(Linear(dim, data_dim))
        self.seq = Sequential(*seq)

    def forward(self, input):
        data = self.seq(input)
        return data


class CTGANSynthesizer(BaseSynthesizer):
    """Conditional Table GAN Synthesizer.

    This is the core class of the CTGAN project, where the different components
    are orchestrated together.
    For more details about the process, please check the [Modeling Tabular data using
    Conditional GAN](https://arxiv.org/abs/1907.00503) paper.
    Args:
        embedding_dim (int):
            Size of the random sample passed to the Generator. Defaults to 128.
        generator_dim (tuple or list of ints):
            Size of the output samples for each one of the Residuals. A Residual Layer
            will be created for each one of the values provided. Defaults to (256, 256).
        discriminator_dim (tuple or list of ints):
            Size of the output samples for each one of the Discriminator Layers. A Linear Layer
            will be created for each one of the values provided. Defaults to (256, 256).
        generator_lr (float):
            Learning rate for the generator. Defaults to 2e-4.
        generator_decay (float):
            Generator weight decay for the Adam Optimizer. Defaults to 1e-6.
        discriminator_lr (float):
            Learning rate for the discriminator. Defaults to 2e-4.
        discriminator_decay (float):
            Discriminator weight decay for the Adam Optimizer. Defaults to 1e-6.
        batch_size (int):
            Number of data samples to process in each step.
        discriminator_steps (int):
            Number of discriminator updates to do for each generator update.
            From the WGAN paper: https://arxiv.org/abs/1701.07875. WGAN paper
            default is 5. Default used is 1 to match original CTGAN implementation.
        verbose (boolean):
            Whether to have print statements for progress results. Defaults to ``False``.
        epochs (int):
            Number of training epochs. Defaults to 300.
        pac (int):
            Number of samples to group together when applying the discriminator.
            Defaults to 10.
        cuda (bool):
            Whether to attempt to use cuda for GPU computation.
            If this is False or CUDA is not available, CPU will be used.
            Defaults to ``True``.
    """

    def __init__(self, embedding_dim=128, generator_dim=(256, 256), discriminator_dim=(256, 256),
                 generator_lr=2e-4, generator_decay=1e-6, discriminator_lr=2e-4,
                 discriminator_decay=1e-6, batch_size=500, discriminator_steps=1,
                 verbose=False, epochs=300, pac=10, cuda=True):

        assert batch_size % 2 == 0

        self._embedding_dim = embedding_dim
        self._generator_dim = generator_dim
        self._discriminator_dim = discriminator_dim

        self._generator_lr = generator_lr
        self._generator_decay = generator_decay
        self._discriminator_lr = discriminator_lr
        self._discriminator_decay = discriminator_decay

        self._batch_size = batch_size
        self._discriminator_steps = discriminator_steps
        self._verbose = verbose
        self._epochs = epochs
        self.pac = pac

        if not cuda or not torch.cuda.is_available():
            device = 'cpu'
        elif isinstance(cuda, str):
            device = cuda
        else:
            device = 'cuda'

        self._device = torch.device(device)

        self._transformer = None
        self._data_sampler = None
        self._generator = None

    @staticmethod
    def _gumbel_softmax(logits, tau=1, hard=False, eps=1e-10, dim=-1):
        """Deals with the instability of the gumbel_softmax for older versions of torch.

        For more details about the issue:
        https://drive.google.com/file/d/1AA5wPfZ1kquaRtVruCd6BiYZGcDeNxyP/view?usp=sharing
        Args:
            logits:
                […, num_features] unnormalized log probabilities
            tau:
                non-negative scalar temperature
            hard:
                if True, the returned samples will be discretized as one-hot vectors,
                but will be differentiated as if it is the soft sample in autograd
            dim (int):
                a dimension along which softmax will be computed. Default: -1.
        Returns:
            Sampled tensor of same shape as logits from the Gumbel-Softmax distribution.
        """
        if version.parse(torch.__version__) < version.parse("1.2.0"):
            for i in range(10):
                transformed = functional.gumbel_softmax(logits, tau=tau, hard=hard,
                                                        eps=eps, dim=dim)
                if not torch.isnan(transformed).any():
                    return transformed
            raise ValueError("gumbel_softmax returning NaN.")

        return functional.gumbel_softmax(logits, tau=tau, hard=hard, eps=eps, dim=dim)

    def _apply_activate(self, data):
        """Apply proper activation function to the output of the generator."""
        data_t = []
        st = 0
        for transformer in self._transformer.transformers:
            if transformer.is_continuous:
                ed = st + transformer.output_width
                data_t.append(torch.tanh(data[:, st:ed]))
                st = ed
            elif transformer.is_categorical:
                ed = st + transformer.output_width
                transformed = self._gumbel_softmax(data[:, st:ed], tau=0.2)
                data_t.append(transformed)
                st = ed

        return torch.cat(data_t, dim=1)

    def _cond_loss(self, data, c, m):
        """Compute the cross entropy loss on the fixed discrete column."""
        loss = []
        st = 0
        st_c = 0
        for t in self._transformer.transformers:
            if not t.is_categorical:
                # not discrete column
                st += t.output_width
            else:
                ed = st + t.output_width
                ed_c = st_c + t.output_width
                tmp = functional.cross_entropy(
                    data[:, st:ed],
                    torch.argmax(c[:, st_c:ed_c], dim=1),
                    reduction='none'
                )
                loss.append(tmp)
                st = ed
                st_c = ed_c

        loss = torch.stack(loss, dim=1)

        return (loss * m).sum() / data.size()[0]

    def _validate_discrete_columns(self, train_data, discrete_columns):
        """Check whether ``discrete_columns`` exists in ``train_data``.

        Args:
            train_data (numpy.ndarray or pandas.DataFrame):
                Training Data. It must be a 2-dimensional numpy array or a pandas.DataFrame.
            discrete_columns (list-like):
                List of discrete columns to be used to generate the Conditional
                Vector. If ``train_data`` is a Numpy array, this list should
                contain the integer indices of the columns. Otherwise, if it is
                a ``pandas.DataFrame``, this list should contain the column names.
        """
        if isinstance(train_data, pd.DataFrame):
            invalid_columns = set(discrete_columns) - set(train_data.columns)
        elif isinstance(train_data, np.ndarray):
            invalid_columns = []
            for column in discrete_columns:
                if column < 0 or column >= train_data.shape[1]:
                    invalid_columns.append(column)
        else:
            raise TypeError('``train_data`` should be either pd.DataFrame or np.array.')

        if invalid_columns:
            raise ValueError('Invalid columns found: {}'.format(invalid_columns))

    # def fit(self, train_data, discrete_columns=tuple(), epochs=None):
    #     """Fit the CTGAN Synthesizer models to the training data.

    #     Args:
    #         train_data (numpy.ndarray or pandas.DataFrame):
    #             Training Data. It must be a 2-dimensional numpy array or a pandas.DataFrame.
    #         discrete_columns (list-like):
    #             List of discrete columns to be used to generate the Conditional
    #             Vector. If ``train_data`` is a Numpy array, this list should
    #             contain the integer indices of the columns. Otherwise, if it is
    #             a ``pandas.DataFrame``, this list should contain the column names.
    #     """
    #     self._validate_discrete_columns(train_data, discrete_columns)

    #     if epochs is None:
    #         epochs = self._epochs
    #     else:
    #         warnings.warn(
    #             ('`epochs` argument in `fit` method has been deprecated and will be removed '
    #              'in a future version. Please pass `epochs` to the constructor instead'),
    #             DeprecationWarning
    #         )

    #     raise ValueError('Call fit method from DPCTGAN or PATECTGAN')

    # def sample(self, n, condition_column=None, condition_value=None):
    #     """Sample data similar to the training data.

    #     Choosing a condition_column and condition_value will increase the probability of the
    #     discrete condition_value happening in the condition_column.
    #     Args:
    #         n (int):
    #             Number of rows to sample.
    #         condition_column (string):
    #             Name of a discrete column.
    #         condition_value (string):
    #             Name of the category in the condition_column which we wish to increase the
    #             probability of happening.
    #     Returns:
    #         numpy.ndarray or pandas.DataFrame
    #     """
    #     if condition_column is not None and condition_value is not None:
    #         condition_info = self._transformer.convert_column_name_value_to_id(
    #             condition_column, condition_value)
    #         global_condition_vec = self._data_sampler.generate_cond_from_condition_column_info(
    #             condition_info, self._batch_size)
    #     else:
    #         global_condition_vec = None

    #     steps = n // self._batch_size + 1
    #     data = []
    #     for i in range(steps):
    #         mean = torch.zeros(self._batch_size, self._embedding_dim)
    #         std = mean + 1
    #         fakez = torch.normal(mean=mean, std=std).to(self._device)

    #         if global_condition_vec is not None:
    #             condvec = global_condition_vec.copy()
    #         else:
    #             condvec = self._data_sampler.sample_original_condvec(self._batch_size)

    #         if condvec is None:
    #             pass
    #         else:
    #             c1 = condvec
    #             c1 = torch.from_numpy(c1).to(self._device)
    #             fakez = torch.cat([fakez, c1], dim=1)

    #         fake = self._generator(fakez)
    #         fakeact = self._apply_activate(fake)
    #         data.append(fakeact.detach().cpu().numpy())

    #     data = np.concatenate(data, axis=0)
    #     data = data[:n]

    #     return self._transformer.inverse_transform(data)

    # def set_device(self, device):
    #     self._device = device
        # if self._generator is not None:
        #     self._generator.to(self._device)



import pandas as pd
import numpy as np
import math


class DPcovariance:

    # Implementation is based off of https://github.com/privacytoolsproject/PSI-Library

    def __init__(self, n, cols, rng, global_eps, epsilon_dist=None, alpha=0.05):

        # TODO finish adding functionality for intercept
        intercept = False

        # The following variables are for different ways of setting up the epsilon value for DP covariance calculation
        # There is infrastructure for them, but we're currently choosing not to expose them.
        epsilon = None
        accuracy = None
        impute_rng = None
        accuracy_vals = None

        self.num_rows = n
        self.columns = cols

        self.intercept = intercept
        self.alpha = alpha

        self.rng = check_range(rng)
        self.sens = covariance_sensitivity(n, rng, intercept)

        if impute_rng is None:
            self.imputeRng = rng
        else:
            self.imputeRng = impute_rng

        if self.intercept:
            self.columns = ["intercept"] + self.columns
        else:
            self.columns = self.columns

        s = len(self.columns)
        output_length = (np.zeros((s, s))[np.tril_indices(s)]).size
        # Distribute epsilon across all covariances that will be calculated
        if epsilon is not None:
            self.epsilon = check_epsilon(epsilon, expected_length=output_length)
            self.globalEps = sum(self.epsilon)
        # Option 2: Enter global epsilon value and vector of percentages specifying how to split global
        # epsilon between covariance calculations.
        elif global_eps is not None and epsilon_dist is not None:
            self.globalEps = check_global_epsilon(global_eps)
            self.epsilonDist = check_epsilon_dist(epsilon_dist, output_length)
            self.epsilon = distribute_epsilon(self.globalEps, epsilon_dist=epsilon_dist)
            self.accuracyVals = laplace_get_accuracy(self.sens, self.epsilon, self.alpha)
        # Option 3: Only enter global epsilon, and have it be split evenly between covariance calculations.
        elif global_eps is not None:
            self.globalEps = check_global_epsilon(global_eps)
            self.epsilon = distribute_epsilon(self.globalEps, n_calcs=output_length)
            self.accuracyVals = laplace_get_accuracy(self.sens, self.epsilon, self.alpha)
        # Option 4: Enter an accuracy value instead of an epsilon, and calculate
        # individual epsilons with this accuracy.
        elif accuracy is not None:
            self.accuracy = check_accuracy(accuracy)
            self.epsilon = laplace_get_epsilon(self.sens, self.accuracy, self.alpha)
            self.globalEps = sum(self.epsilon)
        # Option 5: Enter vector of accuracy values, and calculate ith epsilon value from ith accuracy value
        elif accuracy_vals is not None:
            self.accuracyVals = check_accuracy_vals(accuracy_vals, output_length)
            self.epsilon = laplace_get_epsilon(self.sens, self.accuracyVals, self.alpha)
            self.globalEps = sum(self.epsilon)

    def make_covar_symmetric(self, covar):
        """
        Converts unique private covariances into symmetric matrix

        Args:
            covar (???): differentially privately release of elements in lower triangle of covariance matrix
        Returns:
            A symmetric differentially private covariance matrix (numpy array)
        """
        n = len(self.columns)
        indices = np.triu_indices(n)
        m = np.zeros((n, n))
        m[indices] = covar
        m = m.T
        m = np.tril(m) + np.triu(m.T, 1)
        df = pd.DataFrame(m, columns=self.columns, index=self.columns)
        return df

    def release(self, data):
        new_data = censor_data(data[self.columns], self.rng)
        new_data = fill_missing(new_data, impute_rng=self.imputeRng)

        # TODO: add intercept functionality
        def covar(x, intercept=False):
            if intercept:
                pass  # TODO: Find python equivalent for the following R code: `x < - cbind(1, x)`
            covariance = np.cov(x)
            return list(covariance[np.tril_indices(covariance.shape[0])])

        def q_lap_iter(p, mu=0, b=1):
            for i in range(len(p)):
                p[i] = q_lap(p[i], mu, b[i])
            return p

        def q_lap(elem, mu=0, b=1):
            if elem < 0.5:
                return mu + b * np.log(2 * elem)
            else:
                return mu - b * np.log(2 - 2 * elem)

        def dp_noise(n, noise_scale):
            u = np.random.uniform(size=n)
            return q_lap_iter(u, b=noise_scale)

        true_val = covar(data.values.T, self.intercept)
        scale = self.sens / self.epsilon
        val = np.array(true_val) + dp_noise(n=len(true_val), noise_scale=scale)
        return list(val)

    # TODO: this implementation only works for one dependent variable right now
    def get_linear_regression(self, data, x_names, y_name, intercept=False):
        """
        Takes in data, lists of feature names and target names, and whether or not
        we should calculate a y-intercept; and returns a DP linear regression model

        Args:
            data (Dataframe): the data that will be used to make the linear regression
            x_names (list): list of names of the features (i.e. independent variables) to use
            y_name (string): name of the target (i.e. dependent variable) to use
            intercept (boolean): true if the lin-reg equation should have a y-intercept, false if not
        Return:
            linear regression model (Dataframe) in the following format:
                Each independent variable gets its own row; there are two columns: 'Estimate' and 'Std. Error'.
                'Estimate' is the calculated coefficient for that row's corresponding independent variable,
                'Std. Error' is self evident.

                Here is an example return value given intercept=FALSE, independent variables 'Height' and 'Volume'
                and dependent variable 'Girth':

                           Estimate    Std. Error
                    Height -0.04548    0.02686
                    Volume  0.19518    0.01041
        """
        covar_matrix = self.make_covar_symmetric(self.release(data))
        return cov_method_lin_reg(covar_matrix, self.num_rows, x_names, y_name, intercept)


def cov_method_lin_reg(release, num_rows, x_names, y_name, intercept=False):
    """
        Takes in a differentially privately released covariance matrix, the number of rows in the
        original data, whether or not a y-intercept should be calculated, a list of
        feature names, and a target name; and returns a DP linear regression model

        Args:
            release (Dataframe): differentially privately released covariance matrix that will be used to
                make the linear regression
            num_rows (int): the number of rows in the original data
            x_names (list): list of names of the features (i.e. independent variables) to use
            y_name (string): name of the target (i.e. dependent variable) to use
            intercept (boolean): true if the lin-reg equation should have a y-intercept, false if not
        Returns:
            linear regression model (Dataframe) in the following format:
                Each independent variable gets its own row; there are two columns: 'Estimate' and 'Std. Error'.
                'Estimate' is the calculated coefficient for that row's corresponding independent variable,
                'Std. Error' is self evident.

                Here is an example return value given intercept=FALSE, independent variables 'Height' and 'Volume'
                and dependent variable 'Girth':

                           Estimate    Std. Error
                    Height -0.04548    0.02686
                    Volume  0.19518    0.01041
        """
    eigenvals, _ = list(np.linalg.eig(release.values))
    if not all([ev != 0 for ev in eigenvals]):
        raise ValueError("Matrix is not invertible")
    elif not all([ev > 0 for ev in eigenvals]):
        raise ValueError("Matrix is not positive definite")
    else:
        # Find locations corresponding to the given x & y names
        loc_vec = [False] * release.shape[0]
        row_labels = release.index.values
        x_loc = []
        y_loc = None
        for index in range(len(row_labels)):
            if row_labels[index] in x_names:
                loc_vec[index] = True
                x_loc.append(index)
            if row_labels[index] == y_name:
                y_loc = index
        if x_loc is None or y_loc is None:
            raise ValueError("Names aren't found in the release")

        # Use a sweep to find the coefficient of the independent variable in
        # the linear regression corresponding to the covariance matrix
        sweep = amsweep(release.values / num_rows, np.array(loc_vec))
        coef = sweep[y_loc, x_loc]

        # Calculate the standard error
        submatrix = release.values[x_loc, :][:, x_loc]
        se = list(map(np.sqrt, sweep[y_loc, y_loc] * np.diag(np.linalg.inv(submatrix))))

        new_x_names = [release.index.values[x_loc[i]] for i in range(len(x_loc))]

        def round_5(elem):
            return round(elem, 5)

        # Round both values to account for floating point error, put in Series
        estimates = pd.Series(map(round_5, coef), index=new_x_names, name="Estimate")
        std_error = pd.Series(map(round_5, se), index=new_x_names, name="Std. Error")

        return pd.DataFrame([estimates, std_error]).transpose()


def check_accuracy_vals(accuracy_vals, expected_length):
    if len(accuracy_vals) != expected_length:
        raise ValueError("Epsilon parameter has improper length")
    else:
        for eps in accuracy_vals:
            if eps <= 0:
                raise ValueError("Privacy parameter epsilon must be a value greater than zero")
    return accuracy_vals


def laplace_get_epsilon(sens, accuracy, alpha=0.05):
    return np.log(1 / alpha) * (sens / accuracy)


def check_accuracy(accuracy):
    if accuracy <= 0:
        raise ValueError("Privacy parameter epsilon must be a value greater than zero")
    return accuracy


def laplace_get_accuracy(sens, epsilon, alpha=0.05):
    return np.log(1 / alpha) * (sens / epsilon)


def distribute_epsilon(global_eps, n_calcs=None, epsilon_dist=None):
    if epsilon_dist is None:
        eps = [global_eps / n_calcs for i in range(n_calcs)]
    else:
        eps = [eps * global_eps for eps in epsilon_dist]
    return eps


def check_epsilon_dist(epsilon_dist, expected_length):
    if len(epsilon_dist) != expected_length:
        raise ValueError("Epsilon parameter has improper length")
    else:
        for eps in epsilon_dist:
            if eps <= 0:
                raise ValueError("All values in epsilonDist must be a value greater than zero")
        if sum(epsilon_dist) != 1.0:
            raise ValueError("All values in epsilonDist must sum to 1")
    return epsilon_dist


def check_epsilon(epsilon, expected_length):
    if len(epsilon) != expected_length:
        raise ValueError("Epsilon parameter has improper length")
    else:
        for eps in epsilon:
            if eps <= 0:
                raise ValueError("(Privacy parameter epsilon must be a value greater than zero")
            elif eps >= 3:
                raise ValueError("This is a higher global value than recommended for most cases")
    return epsilon


def check_global_epsilon(eps):
    if eps <= 0:
        raise ValueError("(Privacy parameter epsilon must be a value greater than zero")
    elif eps >= 3:
        raise ValueError("This is a higher global value than recommended for most cases")
    return eps


def covariance_sensitivity(n, rng, intercept):
    diffs = []
    for i in range(rng.shape[1]):
        diffs.append(rng[i][1] - rng[i][0])
    if intercept:
        diffs = [0] + diffs
    const = 2 / n
    sensitivity = []
    for i in range(len(diffs)):
        for j in range(i, len(diffs)):
            s = const * diffs[i] * diffs[j]
            sensitivity.append(s)
    return np.array(sensitivity)


def check_range(rng):
    rng.columns = list(range(rng.shape[1]))
    for col in range(rng.shape[1]):
        rng[col] = rng[col].sort_values()
    return rng


def fill_missing_1D(x, low, high):
    n_missing = x.isnull().sum()
    u = np.random.uniform(size=n_missing)

    def scale(v):
        return v * (high - low) + low

    u = list(map(scale, u))

    def replace_nan(v):
        if math.isnan(v):
            return u.pop()
        return v

    return x.apply(replace_nan)


def fill_missing(data, impute_rng):
    for i in range(data.shape[1]):
        data[i] = fill_missing_1D(data[i], impute_rng[i][0], impute_rng[i][1])
    return data


def censor(value, low, high):
    if value < low:
        return low
    elif value > high:
        return high
    else:
        return value


def censor_data_1D(x, ll, h):
    def scale(v):
        return censor(v, ll, h)

    return x.apply(scale)


def censor_data(data, rng):
    new_data = data

    new_data.columns = list(range(data.shape[1]))
    rng = check_range(rng)

    for i in range(data.shape[1]):
        data[i] = censor_data_1D(data[i], rng[i][0], rng[i][1])
    return data


def amsweep(g, m):
    """
    Sweeps a covariance matrix to extract regression coefficients.

    Args:
        g (Numpy array): a numeric, symmetric covariance matrix divided by the number of observations in the data
        m (Numpy array): a logical vector of length equal to the number of rows in g
        in which the True values correspond to the x values in the matrix
        and the False values correspond to the y values in the matrix

    Return:
        a matrix with the coefficients from g
    """
    # if m is a vector of all falses, then return g
    if np.array_equal(m, np.full(np.shape(m), False, dtype=bool)):
        return g
    else:
        p = np.shape(g)[
            0
        ]  # number of rows of g (np.shape gives a tuple as (rows, cols), so we index [0])
        rowsm = sum(m)  # sum of logical vector "m" (m must be a (n,) shape np array)

        # if all values of m are True (thus making the sum equal to the length),
        # we take the inverse and then negate all the values
        if p == rowsm:
            h = np.linalg.inv(g)  # inverse of g
            h = np.negative(h)  # negate the sign of all elements
        else:
            k = np.where(m)[0]  # indices where m is True
            kcompl = np.where(m is False)[0]  # indices where m is False

            # separate the elements of g
            # make the type np.matrix so that dimensions are preserved correctly
            g11 = np.matrix(g[k, k])
            g12 = np.matrix(g[k, kcompl])
            g21 = np.transpose(g12)
            g22 = np.matrix(g[kcompl, kcompl])

            # use a try-except to get the inverse of g11
            try:
                h11a = np.linalg.inv(g11)  # try to get the regular inverse
            except BaseException:  # should have LinAlgError (not defined error)
                h11a = np.linalg.pinv(g11)
            h11 = np.negative(h11a)

            # matrix multiplication to get sections of h
            h12 = np.matmul(h11a, g12)
            h21 = np.transpose(h12)
            h22 = g22 - np.matmul(np.matmul(g21, h11a), g12)

            # combine sections of h
            hwo = np.concatenate(
                (np.concatenate((h11, h12), axis=1), np.concatenate((h21, h22), axis=1)), axis=0
            )
            hwo = np.asarray(
                hwo
            )  # convert back to array (from matrix) to avoid weird indexing behavior
            xordering = np.concatenate((k, kcompl), axis=0)  # concatenate k and kcompl
            h = np.zeros((p, p))  # make a pxp array of zeros

            for i in range(p):  # traverse each element as defined by xordering
                for j in range(p):
                    h[xordering[i]][xordering[j]] = hwo[i][
                        j
                    ]  # and replace it with the normal i, j element from hwo

        return h



from .base import CachingColumnTransformer
import warnings

class ChainTransformer(CachingColumnTransformer):
    """Sequentially process a column through multiple transforms.  When reversed,
    the inverse transforms are applied in reverse order.

    :param transforms: A list of ColumnTransformers to apply sequentially.
    """
    def __init__(self, transformers):
        self.transformers = transformers
        super().__init__()
    @property
    def output_type(self):
        return self.transformers[-1].output_type
    @property
    def needs_epsilon(self):
        return any(transformer.needs_epsilon for transformer in self.transformers)
    @property
    def cardinality(self):
        cards = []
        for transformer in self.transformers:
            for c in transformer.cardinality:
                cards.append(c)
        return cards
    @property
    def fit_complete(self):
        return all([t.fit_complete for t in self.transformers])
    def allocate_privacy_budget(self, epsilon, odometer):
        n_with_epsilon = sum([1 for t in self.transformers if t.needs_epsilon])
        if n_with_epsilon == 0:
            return
        elif n_with_epsilon > 1:
            warnings.warn(f"Multiple transformers in chain need epsilon, which is likely wasteful.")
        else:
            for transformer in self.transformers:
                if transformer.needs_epsilon:
                    transformer.allocate_privacy_budget(epsilon / n_with_epsilon, odometer)
    def _fit_finish(self):
        vals = self._fit_vals
        for transformer in self.transformers:
            vals = transformer.fit_transform(vals)
        self._fit_vals = []
        self.output_width = self.transformers[-1].output_width
    def _clear_fit(self):
        for transformer in self.transformers:
            transformer._clear_fit()
        if self.fit_complete:
            self.output_width = self.transformers[-1].output_width
    def _transform(self, val):
        for transformer in self.transformers:
            val = transformer._transform(val)
        return val
    def _inverse_transform(self, val):
        for transformer in reversed(self.transformers):
            val = transformer._inverse_transform(val)
        return val